/** @file PlotCanvasScene.cpp
 * 
 * @author marco corvi
 * @date dec 2010
 * 
 * @brief Scene of a plot
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#include <assert.h>
#include <sstream>

#include <QBitmap>
#include <QMessageBox>
#include <QGraphicsSceneMouseEvent>
#include <QLinearGradient>

#include "PlotCanvasScene.h"

#include "Language.h"
// #include "Extend.h"        // extend values
#include "PlotCanvas.h"
#include "PlotScale.h"   // dimensions
#include "PlotZValue.h"
#ifdef HAS_BACKIMAGE
  #include "BackgroundImage.h"
#endif
#include "ThPointType.h"

#define HUNDRED_METERS 300

#define DO_CP 1 /* do control points */

/** size of the ticks of the pit lines (pixels ?)
 */
#define PERCENT_OFFSET 10 /* one twentieth */
#define MIN_OFFSET 50 /* pixels */

/** border pixels in the plots
 */
#define BORDER 0 // 20


// PLOT CANVAS SCENE

PlotCanvasScene::PlotCanvasScene( int x, int y, int w, int h, PlotCanvas * p ) 
  : QGraphicsScene( x, y, w, h, p )
  , parent( p ) 
  , icon( IconSet::Get() )
  , lexicon( Language::Get() )
  , on_line( false )
  , on_area( false )
  , do_splay( false )
  , units( p->getUnits() )
  , status( p->getStatus() )
  , m_zoom( 0 )
  , input_mode( INPUT_COMMAND )
  , point_type( Therion::THP_CLAY )
  , point_orient( 0 )
  , line_type( Therion::THL_WALL )
  // , the_line( NULL )
  , cur_line( NULL )
  // , cur_line_pt( NULL )
  , area_type( Therion::THA_WATER )
  , cur_area( NULL )
  , cur_area_pt( NULL )
  , moving_pt( NULL )
  , moving_line( NULL )
  , moving_area( NULL )
  , moving_line_pt( NULL )
  , moving_item( NULL )
  #ifdef HAS_BACKIMAGE
    , sketch_back( NULL )
  #endif
{ 
  DBG_CHECK("PlotCanvasScene status %p\n", status );
  ig_stations = new QGraphicsItemGroup;
  ig_centerline = new QGraphicsItemGroup;
  ig_splay    = new QGraphicsItemGroup;
  ig_grid     = new QGraphicsItemGroup;
  ig_grid10   = new QGraphicsItemGroup;
  ig_scalebar = new QGraphicsItemGroup;
  addItem( ig_stations );
  addItem( ig_centerline );
  addItem( ig_splay );
  addItem( ig_grid );
  addItem( ig_grid10 );
  addItem( ig_scalebar );
  #ifdef HAS_BACKIMAGE
    // ig_background = new QGraphicsItemGroup;
    // addItem( ig_background );
    // ig_background -> setZValue(  ZVALUE_BACK );
  #endif
  #ifdef HAS_LRUD
    ig_lrud = new QGraphicsItemGroup;
    addItem( ig_lrud );
  #endif

  // this is not the right time to make the plot
  // makeSurvey();
}

PlotCanvasScene::~PlotCanvasScene()
{
  clearPlot();
  #ifdef HAS_LRUD
    removeItem( ig_lrud );
    delete ig_lrud;
  #endif
  removeItem( ig_scalebar );
  removeItem( ig_grid );
  removeItem( ig_grid10 );
  removeItem( ig_splay );
  removeItem( ig_centerline );
  removeItem( ig_stations );
  delete ig_scalebar;
  delete ig_grid;
  delete ig_grid10;
  delete ig_splay;
  delete ig_centerline;
  delete ig_stations;
  #ifdef HAS_BACKIMAGE
    // removeItem( ig_background );
    // delete ig_background;
  #endif
}

// ----------------------------------------------------------------
// MOUSE EVENTS

void 
PlotCanvasScene::mouseReleaseEvent( QGraphicsSceneMouseEvent * e0 )
{
  QPointF p = e0->scenePos();
  // DBG_CHECK("mouseReleaseEvent %.2f %.2f \n", p.x(), p.y() );
  
  if ( parent->getMode() == MODE_3D )
    return;

  #ifdef DO_CP
  double x1 = p.x();
  double y1 = p.y();
  if ( input_mode == INPUT_LINE ) {
    if ( fabs( x1 - cur_pt_x ) > PT_EPS || fabs( y1 - cur_pt_y ) > PT_EPS ) {
      if ( fabs( x1 - cur_pt_x ) < CP_MAX && fabs( y1 - cur_pt_y ) < CP_MAX ) {
        ThLinePoint * pt = insertLinePoint( cur_pt_x, cur_pt_y, line_type );
        setControlPoints( x1, y1, pt );
      }
    } else {
      insertLinePoint( cur_pt_x, cur_pt_y, line_type );
    }
    return;
  }
  #else 
  if ( input_mode == INPUT_LINE /* && cur_line_pt != NULL */ ) {
    return;
  }
  #endif

  if ( input_mode == INPUT_AREA && cur_area_pt != NULL ) {
    // TODO setControlPoints( e0->x(), e0->y(), cur_area_pt );
    return;
  }
  if ( input_mode == INPUT_COMMAND ) {
    if ( moving_pt ) {
      double s = status->getScale();
      double x = p.x();
      double y = p.y();
      if (    fabs( x - s * moving_pt->x ) < PT_EPS
           && fabs( y - s * moving_pt->y ) < PT_EPS ) {
        PointOptionsWidget dialog( this, moving_pt );
      } else {
        moving_pt->x = x / s;
        moving_pt->y = y / s;
        QGraphicsItem * it = moving_pt->item;
        it->setPos( p );
      }
      removeItem( moving_item );
      moving_pt = NULL;
    } else if ( moving_line ) {
      // fprintf(stderr, "Line Point %.2f -> %.2f, %.2f -> %.2f type %d \n", 
      //    moving_line_pt->x, p.x(), moving_line_pt->y, p.y(), moving_line->Type() );
      double dx = p.x() - moving_line_pt->x;
      double dy = p.y() - moving_line_pt->y;
      if ( fabs(dx) < PT_EPS && fabs(dy) < PT_EPS ) {
        LinePointOptionsWidget dialog( this, moving_line_pt, moving_line );
      } else {
        moving_line_pt->x  += dx;
        moving_line_pt->y  += dy;
        moving_line_pt->x1 += dx;
        moving_line_pt->y1 += dy;
        moving_line_pt->x2 += dx;
        moving_line_pt->y2 += dy;
        Therion::LineType type = moving_line->type();
        std::vector< ThLineSegment * >::iterator sit  = moving_line->segmentBegin();
        std::vector< ThLineSegment * >::iterator send = moving_line->segmentEnd();
        for ( ; sit != send; ++sit ) {
          ThLineSegment * sgm = *sit;
          if ( sgm->pt1 == moving_line_pt || sgm->pt2 == moving_line_pt ) {
            QGraphicsLineItem * line = sgm->item;
            if ( type == Therion::THL_PIT || type == Therion::THL_CHIMNEY ) {
              // TODO compute brush
              QColor color;
              QPen pen;
              icon->LineStyle( type, pen, color );
              rotatePenBrush( pen, sgm->pt2->x, sgm->pt2->y, sgm->pt1->x, sgm->pt1->y );
              line->setPen( pen );
            }
            line->setLine( sgm->pt1->x, sgm->pt1->y, sgm->pt2->x, sgm->pt2->y );
          }
        }
        QGraphicsItem * it = moving_line_pt->item;
        it->moveBy( dx, dy );
        it->show();
      }
      moving_line = NULL;
      moving_line_pt = NULL;
      removeItem( moving_item );
    } else if ( moving_area ) {
      removeItem( moving_item );
      std::vector< std::pair<QGraphicsPolygonItem *, ThArea2D *> >::iterator ait = areas.begin();
      std::vector< std::pair<QGraphicsPolygonItem *, ThArea2D *> >::iterator ait_end = areas.end();
      for (  ; ait != ait_end; ++ait ) {
        if ( ait->second == moving_area ) {
          break;
        }
      }
      double dx = p.x() - moving_line_pt->x;
      double dy = p.y() - moving_line_pt->y;
      if ( fabs(dx) < PT_EPS && fabs(dy) < PT_EPS ) {
        AreaPointOptionsWidget dialog( this, moving_line_pt, moving_area );
      } else {
        if ( ait->first ) removeItem( ait->first );
        ait->first = NULL;
        moving_line_pt->x  += dx;
        moving_line_pt->y  += dy;
        moving_line_pt->x1 += dx;
        moving_line_pt->y1 += dy;
        moving_line_pt->x2 += dx;
        moving_line_pt->y2 += dy;
        // Therion::AreaType type = moving_area->Type();
        std::vector< ThLineSegment * >::iterator sit  = moving_area->SegmentBegin();
        std::vector< ThLineSegment * >::iterator send = moving_area->SegmentEnd();
        for ( ; sit != send; ++sit ) {
          ThLineSegment * sgm = *sit;
          if ( sgm->pt1 == moving_line_pt || sgm->pt2 == moving_line_pt ) {
            QGraphicsLineItem * line = sgm->item;
            line->setLine( sgm->pt1->x, sgm->pt1->y, sgm->pt2->x, sgm->pt2->y );
          }
        }
        QGraphicsItem * it = moving_line_pt->item;
        it->moveBy( dx, dy );
        it->show();
        QGraphicsPolygonItem * item = moving_area->getItem();
        if ( item ) {
          // this -> removeItem( item );
          moving_area->setItem( NULL );
        }
        if ( ait != ait_end ) ait->first = createAreaPolygon( moving_area );
      }
      moving_area = NULL;
      moving_line_pt = NULL;
    }
  }
}

void 
PlotCanvasScene::mousePressEvent( QGraphicsSceneMouseEvent * e0 )
{
  QPointF p = e0->scenePos();
  // DBG_CHECK("mousePressEvent %.2f %.2f \n", p.x(), p.y() );
  // fprintf(stderr, "mousePressEvent %.2f %.2f \n", p.x(), p.y() );

  if ( parent->getMode() == MODE_3D )
    return;

  double x0 = p.x();
  double y0 = p.y();
  if ( input_mode == INPUT_LINE ) {
    #ifdef DO_CP
      cur_pt_x = x0;
      cur_pt_y = y0;
    #else
      cur_line_pt = insertLinePoint( x0, y0, line_type );
    #endif
  } else if ( input_mode == INPUT_AREA ) {
    cur_area_pt = insertAreaPoint( x0, y0, area_type );
  } else if ( input_mode == INPUT_POINT ) {  // THERION POINTs
    switch ( point_type ) {
      case Therion::THP_LABEL:
      case Therion::THP_CONTINUATION:
        {
          label_text = "";
          const char * caption = "";
          if ( point_type == Therion::THP_LABEL ) {
            caption = lexicon("qtopo_label");
          } else if ( point_type == Therion::THP_CONTINUATION ) {
            caption = lexicon("qtopo_continuation");
          }
          LabelWidget text( this, caption );
          if ( label_text.size() > 0 ) {
            // std::string option("-text \"");
            // option += label_text;
            // option += "\"";
            insertPoint( x0, y0, point_type, point_orient, label_text.c_str() );
            // insertPointLabel( x0, y0, point_type, label_text.c_str(), point_orient );
          }
        }
        break;
      default:
        insertPoint( x0, y0, point_type, point_orient );
        break;
    }
  } else { // INPUT_COMMAND
    // TherionScrap * scrap = status->getScrap();
    QGraphicsRectItem rect( x0-2, y0-2, 5, 5 );
    QList<QGraphicsItem *> list = collidingItems( &rect );
    // int mode = plot_canvas->getMode();
    if ( list.isEmpty() /* || mode == MODE_CROSS || mode == MODE_HCROSS */ ) {
      DBG_CHECK("Colliding item list is empty\n");
    } else { // if ( mode != MODE_CROSS && mode != MODE_HCROSS ) {
      // fprintf(stderr, "Collisins %d numbers %d cntr-ln %d %d pts %d %d ln_pts %d lns %d ars %d\n",
      //   list.size(), stations.size(), centerline.size(), splay.size(), 
      //   g_points.size(), scrap->pointsSize(),
      //   line_points.size(), scrap->linesSize(),
      //   scrap->areasSize()  );
      for ( QList<QGraphicsItem *>::iterator it = list.begin(); it != list.end(); ++it ) {
        QGraphicsItem * git = *it;
        QGraphicsSimpleTextItem * text = dynamic_cast<QGraphicsSimpleTextItem *>( git );
        if ( text ) {
          PlotPoint* p = getTextPoint( text );
          if ( p ) {
            std::ostringstream oss;
            oss << "-name " << text->text().TO_CHAR();
            if ( ! parent->hasStation( oss.str() ) ) {
              // fprintf(stderr, "inserting station %s at %d %d\n", oss.str().c_str(), p->x0, p->y0 );
              // ThPoint2D * pt =
              insertPoint( p->x0, p->y0, Therion::THP_STATION, 0, oss.str().c_str() );
            } else {
              StationCommentWidget scw( parent, parent, text->text().TO_CHAR() );
            }
            return;
          }
        }
        /* if ( plot_canvas->getMode() == MODE_EXT ) */ 
        QGraphicsLineItem * line = dynamic_cast< QGraphicsLineItem *>( git );
        if ( line ) {
          QGraphicsLineItem * lit = line;
          PlotSegment * s = getSegmentFromLine( lit );
          if ( s && s->block ) {
            ExtendWidget ew( parent, parent, s->block );
            return;
          }
        }
        // check if it is a point
        std::vector< std::pair< QGraphicsItem *, ThPoint2D * > >::iterator pit  = g_points.begin(); 
        std::vector< std::pair< QGraphicsItem *, ThPoint2D * > >::iterator pend = g_points.end(); 
        for ( ; pit != pend; ++pit ) {
          if ( pit->second->type() == Therion::THP_STATION ) continue; // cannot move stations
          if ( pit->first == git ) {
            moving_pt = pit->second;
            double x = moving_pt->x * status->getScale();
            double y = moving_pt->y * status->getScale();
            // fprintf(stderr, "moving point %.2f %.2f \n", moving_pt->x, moving_pt->y );
            moving_item = addEllipse( x-4, y-4, 8, 8, QPen(), QBrush(Qt::red) );
            // moving_item->setBrush( QBrush( Qt::red ) );
            moving_item->setZValue( ZVALUE_MOVING_PT );
            moving_item->show();
            // moving_item->setVisible( true );
            // this -> addItem( moving_item );
            assert( moving_pt->item == git );
            return;
          }
        }
        // line points
        std::vector< std::pair< QGraphicsItem *, ThLine2D * > >::iterator lit  = line_points.begin(); 
        std::vector< std::pair< QGraphicsItem *, ThLine2D * > >::iterator lend = line_points.end(); 
        for ( ; lit != lend; ++lit ) {
          if ( lit->first == git ) {
            ThLine2D * lt = lit->second;
            std::vector< ThLinePoint * >::iterator lpit  = lt->pointBegin();
            std::vector< ThLinePoint * >::iterator lpend = lt->pointEnd();
            for ( ; lpit != lpend; ++lpit ) {
              if ( (*lpit)->item == git ) {
                moving_line = lt;
                moving_line_pt = *lpit;
                moving_item = addEllipse( moving_line_pt->x-4, moving_line_pt->y-4, 8, 8, QPen(), QBrush(Qt::red) );
                moving_item->setZValue( ZVALUE_MOVING_PT );
                moving_item->show();
                return;
              }
            }
          }
        }
        // area points FIXME TODO
        std::vector< std::pair< QGraphicsItem *, ThArea2D * > >::iterator ait  = area_points.begin(); 
        std::vector< std::pair< QGraphicsItem *, ThArea2D * > >::iterator aend = area_points.end(); 
        for ( ; ait != aend; ++ait ) {
          if ( ait->first == git ) {
            ThArea2D * ar = ait->second;
            std::vector< ThLinePoint * >::iterator apit  = ar->Begin();
            std::vector< ThLinePoint * >::iterator apend = ar->End();
            for ( ; apit != apend; ++apit ) {
              if ( (*apit)->item == git ) {
                moving_area = ar;
                moving_line_pt = *apit;
                moving_item = addEllipse( moving_line_pt->x-4, moving_line_pt->y-4, 8, 8, QPen(), QBrush(Qt::red) );
                moving_item->setZValue( ZVALUE_MOVING_PT );
                moving_item->show();
                return;
              }
            }
          }
        }
      }
    }
  }
}

// ---------------------------------------------------------------
// CURSOR

void 
PlotCanvasScene::setOnLine( bool yes_no ) 
{
  on_line = yes_no;
  if ( input_mode == INPUT_LINE ) {
    parent->setCursor( on_line ? icon->PenDown() : icon->PenUp() );
  }
}


void 
PlotCanvasScene::setOnArea( bool yes_no ) 
{
  on_area = yes_no;
  if ( input_mode == INPUT_AREA ) {
    parent->setCursor( on_area ? icon->PenDown() : icon->PenUp() );
  }
}

void 
PlotCanvasScene::setLineType(int t, bool is_line_mode ) 
{ 
  if ( line_type != (Therion::LineType)t ) {
    line_type = (Therion::LineType)t;
    if ( is_line_mode ) {
      endLine( );
    }
  } else if ( !is_line_mode ) {
    endLine( );
  }
}

void 
PlotCanvasScene::setAreaType(int t, bool is_area_mode ) 
{ 
  if ( area_type != (Therion::AreaType)t ) {
    area_type = (Therion::AreaType)t;
    if ( is_area_mode ) {
      endArea( );
    }
  } else if ( !is_area_mode ) {
    endArea( );
  }
}

// this is called only by CanvasCommandWidget::doOK()
void
PlotCanvasScene::setInputMode( InputMode mode )
{
  DBG_CHECK("setInputMode() %d \n", (int)(mode) );

  input_mode = mode;
  if ( mode == INPUT_POINT ) {
    parent->setMode( mode, Therion::PointName[(int)point_type] );
    // cur_line_pt = NULL;
    finishArea();
    cur_area_pt = NULL;
  } else if ( mode == INPUT_LINE ) {
    parent->setMode( mode, Therion::LineName[(int)line_type] );;
    finishArea();
    cur_area_pt = NULL;
  } else if ( mode == INPUT_AREA ) {
    if ( cur_area && cur_area->Type() != area_type ) { // if area type has changed finish area
      finishArea();
    }
    parent->setMode( mode, Therion::AreaName[(int)area_type] );;
    // cur_line_pt = NULL;
  } else { // INPUT_COMMAND
    parent->setMode( mode );
    // cur_line_pt = NULL;
    finishArea();
    cur_area_pt = NULL;
  }
}


// this is called by the PlotCanvas cstr 
//                   and PlotCanvas::redrawPlot()
void
PlotCanvasScene::setOffset()
{
  int offx = MIN_OFFSET;
  int offy = MIN_OFFSET;
  
  int mode = parent->getMode();
  int w = parent->getPlotWidth()  + 2*HUNDRED_METERS; // 200 m
  int h = parent->getPlotHeight() + 2*HUNDRED_METERS;
  if ( mode == MODE_CROSS || mode == MODE_HCROSS ) {
    status->setOffsetX( offx );
    status->setOffsetY( offy );
  } else {
    int dx = - status->getOffsetX(); // previous offset
    int dy = - status->getOffsetY();
    
    // int offx = w/PERCENT_OFFSET;
    // if ( offx < MIN_OFFSET ) offx = MIN_OFFSET;
    // offx -= parent->getPlotXOffset();
    // int offy = h/PERCENT_OFFSET;
    // if ( offy < MIN_OFFSET ) offy = MIN_OFFSET;
    // offy -= parent->getPlotYOffset();
    int offx = HUNDRED_METERS - parent->getPlotXOffset(); 
    int offy = HUNDRED_METERS - parent->getPlotYOffset();
    status->setOffsetX( offx );
    status->setOffsetY( offy );
    // shift drawing by the offset difference
    dx += offx;
    dy += offy;
    TherionScrap * scrap = status->getScrap();
    if ( scrap ) scrap->shiftItems( dx, dy );
    // fprintf(stderr, "plot width %d height %d off %d %d \n", w, h, offx, offy );
  }
  // status->setWidth(  2*offx + parent->getPlotWidth() );
  // status->setHeight( 2*offy + parent->getPlotHeight() );
  status->setWidth(  w );
  status->setHeight( h );
  DBG_CHECK("PlotCanvas::setOffset() to %d %d \n", offx, offy );
  // fprintf(stderr, "PlotCanvas status %d %d off %d %d \n",
  //   status->getWidth(), status->getHeight(),
  //   status->getOffsetX(), status->getOffsetY() );
}

// scan the list of text items ( O(N) )
// was used to find the point where a click occurred
PlotPoint* 
PlotCanvasScene::getTextPoint( QGraphicsSimpleTextItem * text )
{
  ARG_CHECK( text == NULL, NULL );
  DBG_CHECK("getTextPoint() \n");

  std::vector< std::pair<QGraphicsSimpleTextItem*, PlotPoint*> >::iterator end = stations.end();
  std::vector< std::pair<QGraphicsSimpleTextItem*, PlotPoint*> >::iterator it = stations.begin();
  for (  ; it != end; ++it ) {
    if ( text == it->first ) {
      return it->second;
    }
  }
  return NULL;
}

PlotSegment *
PlotCanvasScene::getSegmentFromLine( QGraphicsLineItem * line )
{
  ARG_CHECK( line == NULL, NULL );
  DBG_CHECK("getSegmentFromLine() \n");
  { // centerline segments
    std::vector< std::pair<QGraphicsLineItem *, PlotSegment*> >::iterator end = centerline.end();
    std::vector< std::pair<QGraphicsLineItem *, PlotSegment*> >::iterator it  = centerline.begin();
    for ( ; it != end; ++it ) {
      if ( it->first == line ) {
        return it->second;
      }
    }
  }
  { // splay shots
    std::vector< std::pair<QGraphicsLineItem *, PlotSegment*> >::iterator end = splay.end();
    std::vector< std::pair<QGraphicsLineItem *, PlotSegment*> >::iterator it  = splay.begin();
    for ( ; it != end; ++it ) {
      if ( it->first == line ) {
        return it->second;
      }
    }
  }
  { // other lines
  }
  return NULL;
}


ThPoint2D *
PlotCanvasScene::makePoint( double x, double y, Therion::PointType type,
                            bool add_offset )
{
  double s = status->getScale();
  if ( add_offset ) {
    x += status->getOffsetX() * PT_FACTOR;
    y += status->getOffsetY() * PT_FACTOR;
  }
  double x0 = x / s;
  double y0 = y / s;
  // fprintf(stderr, "makePoint type %d: %.2f %.2f -> %.2f %.2f\n", type, x, y, x0, y0 );
  return new ThPoint2D( x0, y0, type );
}

ThPoint2D *
PlotCanvasScene::insertPointLabel( double x, double y,
                                   Therion::PointType type,
				   const char * text,
                                   int orient )
{
  DBG_CHECK("insertPointLabel X %.2f Y %.2f type %d text %s\n", x, y, type, text );
  ThPoint2D * pt = makePoint( x, y, type );
  switch ( type ) {
    // case Therion::THP_AIR_DRAUGHT:
    // case Therion::THP_WATER_FLOW:
    // case Therion::THP_ENTRANCE:
    case Therion::THP_LABEL:
      pt->setOrientation( orient );
      break;
    default:
      break;
  }
  pt->setText( text );
  TherionScrap * scrap = status->getScrap();
  scrap->pointsAdd( pt );
  status->addUndo( UNDO_POINT );
  parent->turnButtonsOnOff( status->hasUndo() );

  if ( type == Therion::THP_LABEL ) {
    // createLabelPoint( scrap->pointsBack(), option );
    createLabelPoint( scrap->pointsBack(), text, orient );
  } else if ( type == Therion::THP_CONTINUATION ) {
    createLabelPoint( scrap->pointsBack(), "?", 0 ); 
  } else {
    createPoint( scrap->pointsBack() );
  }
  return pt;
}

ThPoint2D *
PlotCanvasScene::insertPoint( double x, double y,
                              Therion::PointType type,
                              int orient,
                              const char * option,
                              bool add_offset )
{
  DBG_CHECK("insertPoint X %.2f Y %.2f type %d option %s\n", x, y, type, option );

  if ( type == Therion::THP_STATION ) {
    ThPoint2D * pt = parent->insertPointStation( x, y, option );
    createPoint( pt );
    return pt;
  }

  if ( type == Therion::THP_LABEL || type == Therion::THP_CONTINUATION ) {
    return insertPointLabel( x, y, type, option, orient );
  }

  // double s = status->getScale();
  // if ( add_offset ) {
  //   x += status->getOffsetX() * PT_FACTOR;
  //   y += status->getOffsetY() * PT_FACTOR;
  // }
  // double x0 = x / s;
  // double y0 = y / s;
  ThPoint2D * pt = makePoint( x, y, type, add_offset );
  switch ( type ) {
    case Therion::THP_AIR_DRAUGHT:
    case Therion::THP_WATER_FLOW:
    case Therion::THP_ENTRANCE:
    case Therion::THP_LABEL: // this never used
      pt->setOrientation( orient );
      break;
    default:
      break;
  }
  if ( option && strlen( option ) > 0 ) {
    pt->setOption( option );
  }
  TherionScrap * scrap = status->getScrap();
  scrap->pointsAdd( pt );
  status->addUndo( UNDO_POINT );
  parent->turnButtonsOnOff( status->hasUndo() );

  createPoint( pt /* scrap->pointsBack() */ );
  return pt;
}

void 
PlotCanvasScene::dropPoint( ThPoint2D * pt ) 
{
  QGraphicsItem * pit = g_points.back().first;
  assert( pit == pt->item );
  this -> removeItem( pt->item );
  g_points.pop_back();
  delete pt;
}

void
PlotCanvasScene::dropLinePoint()
{
  // fprintf(stderr, "dropLinePoint %p pts %d %d sgms %d %d\n",
  //   cur_line, cur_line->pointSize(), line_points.size(),
  //   cur_line->segmentSize(), line_segments.size() );
  ThLinePoint * pt = cur_line->lastPoint();
  ThLineSegment * sgm = cur_line->lastSegment();
  QGraphicsLineItem * lit = line_segments.back().first;
  assert( lit == sgm->item );
  line_segments.pop_back();
  this -> removeItem( lit );

  QGraphicsItem * pit = line_points.back().first;
  assert( pit == pt->item );
  line_points.pop_back();
  this -> removeItem( pit );
  cur_line->dropLast();
  // delete pt;
  // delete sgm;
  // cur_line->pts.pop_back();
  // cur_line->sgms.pop_back(); 
}

void
PlotCanvasScene::dropAreaPoint()
{
  // fprintf(stderr, "dropAreaPoint %p pts %d %d sgms %d %d\n",
  //   cur_area, cur_area->Size(), area_points.size(),
  //   cur_area->SegmentSize(), area_segments.size() );
  ThLinePoint * pt = cur_area->LastPoint();
  ThLineSegment * sgm = cur_area->LastSegment();
  QGraphicsLineItem * lit = area_segments.back().first;
  assert( lit == sgm->item );
  area_segments.pop_back();
  this -> removeItem( lit );

  QGraphicsItem * pit = area_points.back().first;
  assert( pit == pt->item );
  area_points.pop_back();
  this -> removeItem( pit );
  cur_area->DropLast();
  // delete pt;
  // delete sgm;
  // cur_area->pts.pop_back();
  // cur_area->sgms.pop_back(); 
}

void
PlotCanvasScene::dropLine()
{
  ThLinePoint * pt = cur_line->lastPoint();
  QGraphicsItem * pit = line_points.back().first;
  assert( pit == pt->item);
  line_points.pop_back();
  this -> removeItem( pit );
  // delete pt;
  // cur_line->pts.pop_back(); // this is important
  delete cur_line;
  cur_line = NULL;
  setOnLine( false );
}

void
PlotCanvasScene::dropArea()
{
  ThLinePoint * pt = cur_area->LastPoint();
  QGraphicsItem * pit = area_points.back().first;
  assert( pit == pt->item);
  area_points.pop_back();
  this -> removeItem( pit );
  // delete pt;
  // cur_area->pts.pop_back(); // this is important
  delete cur_area;
  cur_area = NULL;
  setOnArea( false );
}

void
PlotCanvasScene::setControlPoints( double x, double y, ThLinePoint * pt, bool add_offset )
{
  if ( add_offset ) {
    x += status->getOffsetX() * PT_FACTOR;
    y += status->getOffsetY() * PT_FACTOR;
  }
  pt->setControls( x, y );
}

void
PlotCanvasScene::setControlPoint1( double x, double y, ThLinePoint * pt, bool add_offset )
{
  if ( add_offset ) {
    x += status->getOffsetX() * PT_FACTOR;
    y += status->getOffsetY() * PT_FACTOR;
  }
  pt->setControl1( x, y );
}

void
PlotCanvasScene::setControlPoint2( double x, double y, ThLinePoint * pt, bool add_offset )
{
  if ( add_offset ) {
    x += status->getOffsetX() * PT_FACTOR;
    y += status->getOffsetY() * PT_FACTOR;
  }
  pt->setControl2( x, y );
}




ThLinePoint *
PlotCanvasScene::insertLinePoint( double x, double y, Therion::LineType type, bool add_offset )
{
  DBG_CHECK("insertLinePoint X %.2f Y %.2f Line type %s on-line %d\n", 
    x, y, PlotExport::ThLineName[(int)type], (int)on_line );
  TherionScrap * scrap = status->getScrap();

  ThLinePoint * ret = NULL;
  if ( add_offset ) {
    x += status->getOffsetX() * PT_FACTOR;
    y += status->getOffsetY() * PT_FACTOR;
  }

  // if ( cur_line == NULL ) {
  if ( on_line == false ) {
    DBG_CHECK("creating a new line \n");
    cur_line = new ThLine2D( type );
    scrap->linesAdd( cur_line );
    setOnLine( true );
    // fprintf(stderr, "new line %p\n", cur_line );
  } else {
    // fprintf(stderr, "cur line %p pts %d sgms %d\n",
    //   cur_line, cur_line->pointSize(), cur_line->segmentSize() );
  }
  size_t size = cur_line->pointSize();
  if ( size > 0 ) {
    ThLinePoint * pt     =  cur_line->operator[](0);
    ThLinePoint * ptlast =  cur_line->operator[](size-1);
    if ( fabs( pt->x - x ) < PT_EPS && fabs( pt->y - y ) < PT_EPS ) {
      // double x1 = pt->x;
      // double y1 = pt->y;
      // QGraphicsItem * it = pt->item;
      // x = x1*s;
      // y = y1*s;
      // FIX-2 ret = cur_line->Add( x1, y1, it );
      QGraphicsLineItem * sgm = NULL;
      createLineSegmentTo( pt, type, size, NULL, &sgm );
      cur_line->closeLine( sgm );
      // fprintf(stderr, "closed line %p pts %d sgms %d\n",
      //   cur_line, cur_line->pointSize(), cur_line->segmentSize());
      setOnLine( false );
      cur_line = NULL;
      status->addUndo( UNDO_CLOSELINE );
      parent->turnButtonsOnOff( status->hasUndo() );
    } else if ( fabs( ptlast->x - x ) < PT_EPS && fabs( ptlast->y - y ) < PT_EPS ) {
      DBG_CHECK("insertLinePoint end open line\n" );
      endLine();
      // dumpUndoList( status->undo );
    } else {
      DBG_CHECK("insertLinepoint() line size before insert %d\n", size);
      QGraphicsItem * pt = NULL;
      QGraphicsLineItem * sgm = NULL;;
      ret = new ThLinePoint( x, y, NULL );
      createLineSegmentTo( ret, type, size, &pt, &sgm );
      cur_line->addLinePoint( ret, pt, sgm );
      status->addUndo( UNDO_LINEPOINT );
      parent->turnButtonsOnOff( status->hasUndo() );
    }
  } else {
    DBG_CHECK("insertLinepoint() first line point\n");
    QGraphicsItem * pt = NULL;
    ret = new ThLinePoint( x, y, NULL );
    createLineSegmentTo( ret, type, size, &pt, NULL ); 
    cur_line->addLinePoint( ret, pt, NULL );
    status->addUndo( UNDO_LINEPOINT );
    parent->turnButtonsOnOff( status->hasUndo() );
    // dumpUndoList( status->undo );
  }
  return ret;
}

void
PlotCanvasScene::finishArea( )
{
  if ( cur_area == NULL ) return;
  size_t size = cur_area->Size();
  if ( size <= 2 ) { 
    // delete area
    std::vector< std::pair<QGraphicsItem *, ThArea2D *> >::iterator pt  = area_points.begin();
    while ( pt != area_points.end() ) {
      if ( pt->second == cur_area ) {
        removeItem( pt->first );
        pt = area_points.erase( pt );
      } else {
        ++pt;
      }
    }
    std::vector< std::pair<QGraphicsLineItem *, ThArea2D *> >::iterator ln  = area_segments.begin();
    while ( ln != area_segments.end() ) {
      if ( ln->second == cur_area ) {
        removeItem( ln->first );
        ln = area_segments.erase( ln );
      } else {
        ++ln;
      }
    }
    delete cur_area;
    cur_area = NULL;
    return;
  }
  ThLinePoint * pt =  cur_area->operator[](0);
  double x1 = pt->x;
  double y1 = pt->y;
  QGraphicsLineItem * sgm = NULL;
  createAreaSegmentTo( x1, y1, cur_area->Type(), size, NULL, &sgm );
  cur_area->Finish( sgm );
  // fprintf(stderr, "finish area %p pts %d sgms %d\n", cur_area, cur_area->Size(), cur_area->SegmentSize());
  setOnArea( false );
  QGraphicsPolygonItem * item = createAreaPolygon( cur_area );
  areas.push_back( std::pair< QGraphicsPolygonItem *, ThArea2D * >( item, cur_area ) );
  cur_area = NULL;
  status->addUndo( UNDO_FINISHAREA );
  parent->turnButtonsOnOff( status->hasUndo() );
}


QGraphicsPolygonItem * 
PlotCanvasScene::createAreaPolygon( ThArea2D * area )
{
  QVector< QPointF > contour;
  for ( size_t k = 0; k < area->Size(); ++k ) {
    ThLinePoint * pt = area->operator[](k);
    contour.append( QPointF(pt->x, pt->y) );
  }
  QPolygonF polygon( contour );
  QPen   pen;
  QBrush brush;
  QColor color;
  icon->AreaStyle( area->Type(), color, pen, brush );
  QGraphicsPolygonItem * item = addPolygon( polygon, pen, brush );
  item->setZValue( ZVALUE_AREA );
  area->setItem( item );
  return item;
}

ThLinePoint *
PlotCanvasScene::insertAreaPoint( double x, double y, Therion::AreaType type, bool add_offset )
{
  DBG_CHECK("insertAreaPoint X %.2f Y %.2f Line type %s on-area %d\n", 
    x, y, PlotExport::ThAreaName[(int)type], (int)on_area );
  TherionScrap * scrap = status->getScrap();

  ThLinePoint * ret = NULL;
  if ( add_offset ) {
    x += status->getOffsetX() * PT_FACTOR;
    y += status->getOffsetY() * PT_FACTOR;
  }

  // if ( cur_area == NULL ) {
  if ( on_area == false ) {
    DBG_CHECK("creating a new area \n");
    cur_area = new ThArea2D( type );
    scrap->areasAdd( cur_area );
    setOnArea( true );
    // fprintf(stderr, "new area %p\n", cur_area );
  } else {
    // fprintf(stderr, "cur area %p pts %d sgms %d\n",
    //   cur_area, cur_area->Size(), cur_area->SegmentSize() );
  }
  size_t size = cur_area->Size();
  if ( size > 0 ) {
    ThLinePoint * pt     =  cur_area->operator[](0);
    ThLinePoint * ptlast =  cur_area->operator[](size-1);
    if ( fabs( pt->x - x ) < PT_EPS && fabs( pt->y - y ) < PT_EPS ) {
      finishArea();
    } else if ( fabs( ptlast->x - x ) < PT_EPS && fabs( ptlast->y - y ) < PT_EPS ) {
      DBG_CHECK("insertAreaPoint cannot end open area\n" );
      // endArea();
      // dumpUndoList( status->undo );
    } else {
      DBG_CHECK("insertAreapoint() area size before insert %d\n", size);
      QGraphicsItem * pt = NULL;
      QGraphicsLineItem * sgm = NULL;;
      createAreaSegmentTo( x, y, type, size, &pt, &sgm );
      ret = cur_area->addAreaPoint( x, y, pt, sgm );
      status->addUndo( UNDO_AREAPOINT );
      parent->turnButtonsOnOff( status->hasUndo() );
    }
  } else {
    DBG_CHECK("insertAreapoint() first area point\n");
    QGraphicsItem * pt = NULL;
    createAreaSegmentTo( x, y, type, size, &pt, NULL );
    ret = cur_area->addAreaPoint( x, y, pt, NULL );
    status->addUndo( UNDO_AREAPOINT );
    parent->turnButtonsOnOff( status->hasUndo() );
    // dumpUndoList( status->undo );
  }
  return ret;
}
    
void 
PlotCanvasScene::endLine( bool do_assert ) 
{
  if ( cur_line ) {
    DBG_CHECK("endLine() cur_line %p\n", cur_line );
    setOnLine( false );
    cur_line = NULL;
    status->addUndo( UNDO_ENDLINE );
  } else if ( do_assert ) {
    assert( on_line == false );
  }
}
    
void 
PlotCanvasScene::endArea( bool do_assert ) 
{
  if ( cur_area ) {
    DBG_CHECK("endArea() cur_area %p\n", cur_area );
    setOnArea( false );
    cur_area = NULL;
    status->addUndo( UNDO_FINISHAREA );
  } else if ( do_assert ) {
    assert( on_area == false );
  }
}


void
PlotCanvasScene::uncloseLine( ThLine2D * line )
{
  TherionScrap * scrap = status->getScrap();
  cur_line = line;
  QGraphicsLineItem * lit = cur_line->lastSegment()->item;
  assert( lit == line_segments.back().first );
  cur_line->unclose();
  this -> removeItem( lit );
  line_segments.pop_back();
  setCurLine( scrap->linesBack() );
  curLine()->unclose( );
  resetLineType( cur_line->type() );
  setInputMode( INPUT_LINE );
  setOnLine( true );
}

void
PlotCanvasScene::unfinishArea( ThArea2D * area )
{
  TherionScrap * scrap = status->getScrap();
  cur_area = area;
  QGraphicsLineItem * ait = cur_area->LastSegment()->item;
  assert( ait == area_segments.back().first );
  // cur_area->Unfinish();
  this -> removeItem( ait );
  QGraphicsPolygonItem * item = area->getItem();
  if ( item ) {
    this -> removeItem( item );
    area->setItem( NULL );
  }
  area_segments.pop_back();
  setCurArea( scrap->areasBack() );
  curArea()->Unfinish( );
  resetAreaType( cur_area->Type() );
  setInputMode( INPUT_AREA );
  setOnArea( true );
}

void
PlotCanvasScene::rotatePenBrush( QPen & pen, double x1, double y1, double x2, double y2 )
{
  QBrush brush = pen.brush();
  QMatrix m;
  double angle = atan2( y2-y1, x2-x1 ) - M_PI/2 ;
  m.translate( x1-10*cos(angle),  y1-10*sin(angle));
  m.rotate( angle * 180 / M_PI );
  brush.setMatrix( m );
  pen.setBrush( brush );
}

void
PlotCanvasScene::createLineSegmentTo( ThLinePoint * line_pt, Therion::LineType type, size_t size,
                                      QGraphicsItem ** ppt, QGraphicsLineItem ** psgm )
{
  if ( cur_line == NULL ) return;
  // double s = status->getScale();
#ifdef PLOT_PAINTER
  ThLinePoint * prev = (size == 0 )? NULL : cur_line->operator[]( size - 1 );

  plot_painter.createLineSegmentTo( prev, line_pt, type, ppt, psgm );
  if ( ppt && *ppt ) {
    this -> addItem( *ppt );
    line_points.push_back( std::pair< QGraphicsItem*, ThLine2D*>(*ppt, cur_line) );
  }
  if ( psgm && *psgm ) {
    this -> addItem( *psgm );
    line_segments.push_back( std::pair< QGraphicsLineItem*, ThLine2D *>( *psgm, cur_line ) );
  }

#else
  double x = line_pt->x;
  double y = line_pt->y;
  DBG_CHECK("createLineSegmentTo() %.2f %.2f \n", x, y );
  // end-point
  if ( ppt ) {
    QColor color = icon->LinePointColor( type );
    QGraphicsItem * pt = NULL;
    if ( type == Therion::THL_ARROW ) {
      if ( size == 0 ) {
        pt = createRoundLinePoint( x, y, color );
      } else {
        int k = 0;
        ThLinePoint * prev = cur_line->operator[]( size - 1 );
        double dx = x - prev->x;
        double dy = y - prev->y;
        if ( fabs(dx) > 2*fabs(dy) ) {
          k = (dx > 0)? 2 : 6;
        } else if ( fabs(dy) > 2*fabs(dx) ) { 
          k = (dy > 0)? 4 : 0;
        } else if ( dx > 0 ) {
          k = (dy > 0)? 3 : 1;
        } else {
          k = (dy > 0)? 5 : 7;
        }
        pt = createPolyPoint( x, y, icon->ArrowEnd(k) );
      }
    } else {
      pt = createRoundLinePoint( x, y, color );
    }
    *ppt = pt;
  }

  // line-segment
  // FIXME use special brushes for pit and chimney
  if ( psgm ) {
    QGraphicsLineItem * sgm = NULL;
    if ( size > 0 ) {
      ThLinePoint * prev = cur_line->operator[]( size - 1 );
      QColor color;
      QPen pen;
      icon->LineStyle( type, pen, color );
      if ( type == Therion::THL_PIT || type == Therion::THL_CHIMNEY ) {
        rotatePenBrush( pen, prev->x, prev->y, x, y );
      }
      sgm = createLineSegment( prev->x, prev->y, x, y, pen );
    }
    *psgm = sgm;
  }
#endif
}


void
PlotCanvasScene::createAreaSegmentTo( double x, double y, Therion::AreaType type, size_t size,
                                      QGraphicsItem ** ppt, QGraphicsLineItem ** psgm )
{
  DBG_CHECK("createAreaSegmentTo() %.2f %.2f \n", x, y );

  if ( cur_area == NULL ) return;
  // double s = status->getScale();

#ifdef PLOT_PAINTER
  ThLinePoint * prev = ( size > 0 )? cur_area->operator[]( size - 1 ) : NULL ;
  plot_painter.createAreaSegmentTo( x, y, prev, type, ppt, psgm );
  if ( ppt && *ppt ) {
    this -> addItem( *ppt );
    area_points.push_back( std::pair< QGraphicsItem*, ThArea2D*>(*ppt,cur_area) );
  }
  if ( psgm && *psgm ) {
    this -> addItem( *psgm );
    area_segments.push_back( std::pair< QGraphicsLineItem*, ThArea2D *>( *psgm, cur_area ) );
  }
#else
  // end-point
  if ( ppt ) {
    QGraphicsItem * pt = NULL;
    if ( type == Therion::THA_WATER ) {
      pt = createRoundAreaPoint( x, y, Qt::blue );
    } else if ( type == Therion::THA_SNOW || type == Therion::THA_ICE ) {
      pt = createRoundAreaPoint( x, y, Qt::gray );
    } else if ( type == Therion::THA_USER ) {
      pt = createRoundAreaPoint( x, y, Qt::green );
    } else {
      pt = createRoundAreaPoint( x, y, Qt::black );
    }
    *ppt = pt;
  }

  // area-segment
  if ( psgm ) {
    QGraphicsLineItem * sgm = NULL;
    if ( size > 0 ) {
      ThLinePoint * prev = cur_area->operator[]( size - 1 );
      QColor color;
      QPen pen;
      QBrush brush;
      icon->AreaStyle( type, color, pen, brush );
      // if ( type == Therion::THA_ICE ) {
      //   rotatePenBrush( pen, prev->x, prev->y, x, y );
      // }
      sgm = createAreaSegment( prev->x, prev->y, x, y, pen );
    }
    *psgm = sgm;
  }
#endif
}

// ===================================================================
// DISPLAY and CLEAR

void
PlotCanvasScene::displayNumbers( )
{
  if ( status->isNumbers() ) {
    ig_stations->show();
  } else {
    ig_stations->hide();
  }
}

void 
PlotCanvasScene::displayGrid( int s ) 
{
  if ( s == PLOT_FRAME_GRID ) {
    ig_scalebar->hide();
    if ( m_zoom > -4 ) {
      ig_grid->show();
    } else {
      ig_grid->hide();
    }
    ig_grid10->show();
  } else if ( s == PLOT_FRAME_BAR ) {
    DBG_CHECK("show scalebar\n");
    ig_grid->hide();
    ig_grid10->hide();
    ig_scalebar->show();
  } else if ( s == PLOT_FRAME_3D ) {
    ig_grid->hide();
    ig_grid10->hide();
    ig_scalebar->hide();
  } else {
    if ( m_zoom > -4 ) {
      ig_grid->show();
    } else {
      ig_grid->hide();
    }
    ig_grid10->show();
    ig_scalebar->show();
  }
}

void
PlotCanvasScene::makeSurvey( )
{
// DBG_CHECK("makeSurvey() mode %d theta %.2f phi %.2f\n", mode, status->getTheta(), status->getPhi() );
  {
    double s = status->getScale();
    int ox = status->getOffsetX();
    int oy = status->getOffsetY();
  
    // int w2 = (int)((plot->x0max + ox*2)*s) + BORDER;
    // int h2 = (int)((plot->y0max + oy*2)*s) + BORDER;
    // status->width  = w2;
    // status->height = h2;
    int w2 = (int)(s * status->getWidth() ) + BORDER;
    int h2 = (int)(s * status->getHeight() ) + BORDER;
    // FIXME canvas->resize( w2, h2 );Q
    setSceneRect( ox, oy, w2, h2 );
  }
  clearSurvey();
/* FIXME needed ?
  if ( parent->getMode() != MODE_3D ) {
    double ls = units.length_factor;
    parent->getFrame()->compouteGrid( s/ls,
                   (int)(status->getWidth() * ls),
                   (int)(status->getHeight() * ls),
                   units.length_unit );
    parent->showGridSpacing();
  }
*/
  double zmax = -10000, zmin=10000;
  if ( parent->getMode() == MODE_3D ) {
    // FIXME is it possible to get zmax/zmin from num ?
    for ( PlotSegment * cs = parent->getPlot()->getSegments(); cs; cs=cs->next ) {
      if ( cs->cs_type == CS_CENTERLINE ) {
        if ( zmax < cs->z0 ) zmax = cs->z0;
        if ( zmin > cs->z0 ) zmin = cs->z0;
        if ( zmax < cs->z1 ) zmax = cs->z1;
        if ( zmin > cs->z1 ) zmin = cs->z1;
      }
    }
    zmax = 196.0/(zmax-zmin);
  }

  for ( PlotSegment * cs = parent->getPlot()->getSegments(); cs; cs=cs->next ) {
    // if ( ! do_splay && cs->cs_type != CS_CENTERLINE ) continue;
    QLineF linef( status->evalToX(cs->x0), status->evalToY(cs->y0),
                  status->evalToX(cs->x1), status->evalToY(cs->y1) );
    QGraphicsLineItem * line = new QGraphicsLineItem( linef );
    this -> addItem( line );
    if ( cs->cs_type == CS_CENTERLINE ) {
      ig_centerline->addToGroup( line );
      centerline.push_back( std::pair<QGraphicsLineItem *, PlotSegment*>(line, cs) );
    } else {
      ig_splay->addToGroup( line );
      splay.push_back( std::pair<QGraphicsLineItem *, PlotSegment*>(line, cs) );
    }

    if ( parent->getMode() == MODE_3D && cs->cs_type == CS_CENTERLINE ) {
      QPen p( icon->PenRed() );
      QLinearGradient g( status->evalToX(cs->x0), status->evalToY(cs->y0),
                         status->evalToX(cs->x1), status->evalToY(cs->y1) );
      int r0 =  196 - (int)(zmax*(cs->z0 - zmin));
      int r1 =  196 - (int)(zmax*(cs->z1 - zmin));
      g.setColorAt( 0.0, QColor( 255, r0, r0, 255 ) );
      g.setColorAt( 1.0, QColor( 255, r1, r1, 255 ) );
      p.setBrush( g );
      p.setWidth( 2 );
      // p.setWidth( (cs->z0 + cs->z1)/10 );
      line->setPen( p );
      line->setZValue( ZVALUE_LINE );
    } else {
      QPen p( (cs->cs_type == CS_CENTERLINE )? icon->PenRed() : icon->PenDarkGray()  );
      if ( cs->block->hasComment() ) { // lines with comments are thick
        // p = (cs->cs_type == CS_CENTERLINE )? icon->DarkRed() : icon->PenBlack();
        p.setWidth( 2 );
      }
      line->setPen( p );
      line->setZValue( ZVALUE_LINE );
    }
  }
  ig_centerline->show();
  if ( do_splay ) {
    ig_splay->show();
  } else {
    ig_splay->hide();
  }
  
  for ( PlotPoint * pt = parent->getPlot()->getPoints(); pt; pt=pt->next ) {
    QGraphicsSimpleTextItem * text = this -> addSimpleText( QString(pt->name) );
    text->setBrush( parent->getList()->hasStationComment(pt->name)? Qt::blue : Qt::darkCyan );
    // text->setColor( parent->GetList()->hasStationComment(pt->name)? Qt::darkBlue : Qt::blue );
    text->setPos( status->evalToX(pt->x0), status->evalToY(pt->y0) );
    text->setZValue( ZVALUE_NUMBER );
    ig_stations->addToGroup( text );
    stations.push_back( std::pair<QGraphicsSimpleTextItem*, PlotPoint*>(text,pt) );
    // fprintf(stderr, "add station %d number %s \n", stations.size(), pt->name );
  }
  displayNumbers();
}

void 
PlotCanvasScene::showPlot()
{
  // printf("showPlot() ...\n");
  makeSurvey();
  TherionScrap * scrap = status->getScrap();
  if ( scrap == NULL ) return;

  if ( parent->getMode() != MODE_3D ) {
    // LINES
    for ( std::vector< ThLine2D * >::iterator lit = scrap->linesBegin(), lend = scrap->linesEnd();
          lit != lend;
          ++lit ) {
      ThLine2D::point_iterator it = (*lit)->pointBegin();
      ThLine2D::point_iterator end = (*lit)->pointEnd();
#ifdef PLOT_PAINTER
      cur_line = *lit;
      ThLinePoint * prev = NULL;
      for ( ; it != end; ++it ) {
        QGraphicsItem * ppt = NULL;
        QGraphicsLineItem * psgm = NULL;
        ThLinePoint * pt = *it;
        plot_painter.createLineSegmentTo( prev, pt, cur_line->type(), &ppt, &psgm );
        if ( ppt != NULL ) {
          this -> addItem( ppt );
          line_points.push_back( std::pair< QGraphicsItem*, ThLine2D*>(ppt, cur_line) );
        }
        if ( psgm != NULL ) {
          this -> addItem( psgm );
          line_segments.push_back( std::pair< QGraphicsLineItem*, ThLine2D *>( psgm, cur_line ) );
        }
        prev = pt;
      }
      cur_line = NULL;
#else
      ThLinePoint * pt = *it;
      QColor color;
      QPen pen;
      icon->LineStyle( (*lit)->type(), pen, color );
      double x = pt->x;
      double y = pt->y;
      QGraphicsItem * item;
      item = createRoundLinePoint( x, y, color ); // FIXME x, y
      for ( ++it; it != end; ++it ) {
        pt = *it;
        if ( (*lit)->type() == Therion::THL_ARROW ) {
          int k = 0;
          double dx = pt->x - x;
          double dy = pt->y - y;
          if ( fabs(dx) > 2*fabs(dy) ) {
            k = (dx > 0)? 2 : 6;
          } else if ( fabs(dy) > 2*fabs(dx) ) { 
            k = (dy > 0)? 4 : 0;
          } else if ( dx > 0 ) {
            k = (dy > 0)? 3 : 1;
          } else {
            k = (dy > 0)? 5 : 7;
          }
          item = createPolyPoint( pt->x, pt->y, icon->ArrowEnd(k) );
        } else {
          item = createRoundLinePoint( pt->x, pt->y, color ); // FIXME x, y
        }
        createLineSegment( x, y, pt->x, pt->y, pen );
        x = pt->x;
        y = pt->y;
      }
#endif
    }

    // POINTS
    for ( std::vector< ThPoint2D * >::iterator it = scrap->pointsBegin(), 
          end = scrap->pointsEnd(); it != end; ++it ) {
      ThPoint2D * pt = *it;
      if ( pt->type() == Therion::THP_LABEL ) {
        createLabelPoint( pt, pt->text(), pt->orientation() );
      } else if ( pt->type() == Therion::THP_CONTINUATION ) {
        createLabelPoint( pt, "?", 0 ); // six spaces followed by '?'
      } else {
        createPoint( pt );
      }
    }
  }
  // FIXME parent->show();
}

void
PlotCanvasScene::makeGrid( int d0, int gx, int gy, int dx, int dy )
{
  QPen p( Qt::lightGray );
  grid_size = gx+gy;
  grid = new QGraphicsLineItem * [ gx+gy ];
  for (int k=0; k<gx; ++k ) {
    int x = d0/2+k*d0;
    grid[k] = addLine( x, 0, x, dy, p );
    if ( k%10 != 5 ) {
      grid[k]->setZValue( ZVALUE_GRID );
      ig_grid->addToGroup( grid[k] );
    } else {
      grid[k]->setPen( QPen(Qt::darkGray) );
      grid[k]->setZValue( ZVALUE_GRID10 );
      ig_grid10->addToGroup( grid[k] );
    }
    // grid[k]->show();
  }
  for (int k=gx; k<gx+gy; ++k ) {
    int y = d0/2+(k-gx)*d0;
    grid[k] = addLine( 0, y, dx, y, p );
    if ( k%10 != 5 ) {
      grid[k]->setZValue( ZVALUE_GRID );
      ig_grid->addToGroup( grid[k] );
    } else {
      grid[k]->setPen( QPen(Qt::darkGray) );
      grid[k]->setZValue( ZVALUE_GRID10 );
      ig_grid10->addToGroup( grid[k] );
    }
    // grid[k]->show();
  }
}

void
PlotCanvasScene::makeScaleBar( int x1, int x2, int y, const char * scale )
{
  x1 += 300;
  x2 += 300;
  y  += 160;
  QPen p( icon->PenBlack() );
  p.setWidth( 5 );
  scale_bar = addLine( x1, y, x2, y, p );
  scale_bar->setZValue( ZVALUE_SCALE );
  scale_bar->show();
  scale_text = addSimpleText( QString(scale) );
  scale_text->setBrush( Qt::black );
  scale_text->setPos( x2, y );
  scale_text->setZValue( ZVALUE_SCALE );
  scale_text->show();
  ig_scalebar -> addToGroup( scale_bar );
  ig_scalebar -> addToGroup( scale_text );
}

#ifndef PLOT_PAINTER
QGraphicsLineItem *
PlotCanvasScene::createLineSegment( double x1, double y1, double x2, double y2, const QPen & pen )
{
  DBG_CHECK("createLineSegment %.2f %.2f - %.2f %.2f\n", x1, y1, x2, y2 );
  QLineF linef( x1, y1, x2, y2 );
  QGraphicsLineItem * line = new QGraphicsLineItem( linef );
  line->setPen( pen );
  line->setZValue( ZVALUE_LINE );
  line->show();
  this -> addItem( line );
  line_segments.push_back( std::pair< QGraphicsLineItem*, ThLine2D *>( line, cur_line ) );
  return line;
}

QGraphicsLineItem *
PlotCanvasScene::createAreaSegment( double x1, double y1, double x2, double y2, const QPen & pen )
{
  DBG_CHECK("createAreaSegment %.2f %.2f - %.2f %.2f\n", x1, y1, x2, y2 );
  QLineF linef( x1, y1, x2, y2 );
  QGraphicsLineItem * line = new QGraphicsLineItem( linef );
  line->setPen( pen );
  line->setZValue( ZVALUE_LINE );
  line->show();
  this -> addItem( line );
  area_segments.push_back( std::pair< QGraphicsLineItem*, ThArea2D *>( line, cur_area ) );
  return line;
}

// used for LINE points
QGraphicsEllipseItem *
PlotCanvasScene::createRoundLinePoint( double x, double y, const QColor & color )
{
  DBG_CHECK("createRoundLinePoint %.2f %.2f\n", x, y );
  QGraphicsEllipseItem * ell = new QGraphicsEllipseItem( x-2, y-2, 4, 4 );
  ell->setPen(  color );
  ell->setBrush( QBrush( color ) );
  ell->setZValue( ZVALUE_LINE_PT );
  ell->show();
  this -> addItem( ell );
  line_points.push_back( std::pair< QGraphicsItem*, ThLine2D*>(ell,cur_line) );
  return ell;
}

// used for AREA points
QGraphicsEllipseItem *
PlotCanvasScene::createRoundAreaPoint( double x, double y, const QColor & color )
{
  DBG_CHECK("createRoundAreaPoint %.2f %.2f\n", x, y );
  QGraphicsEllipseItem * ell = new QGraphicsEllipseItem( x-2, y-2, 4, 4 );
  ell->setBrush( QBrush( color ) );
  ell->setZValue( ZVALUE_LINE_PT );
  ell->show();
  this -> addItem( ell );
  area_points.push_back( std::pair< QGraphicsItem*, ThArea2D*>(ell,cur_area) );
  return ell;
}

// used for ARROW heads
QGraphicsItem *
PlotCanvasScene::createPolyPoint( double x, double y, const QPolygon & pts )
{
  QGraphicsPolygonItem * poly = new QGraphicsPolygonItem( pts );
  poly->setBrush( QBrush( Qt::black ) );
  poly->setPos( x, y );
  poly->setZValue( ZVALUE_LINE_PT );
  poly->show();
  this -> addItem( poly );
  line_points.push_back( std::pair< QGraphicsItem*, ThLine2D*>(poly,cur_line) );
  return poly;
}
#endif

void
PlotCanvasScene::greyOut( TherionScrap * scrap )
{
  // fprintf(stderr, "greying out %s pst %d lns %d \n", scrap->getScrapName(), 
  //   scrap->pointsSize(), scrap->linesSize() );  
  assert( scrap->getGrey() == NULL );
  QGraphicsItemGroup * group = new QGraphicsItemGroup;
  this -> addItem( group );
  scrap ->setGrey( group );
  std::vector< ThLine2D * >::const_iterator lit = scrap->linesBegin();
  std::vector< ThLine2D * >::const_iterator lit_end = scrap->linesEnd();
  for ( ; lit != lit_end; ++ lit ) {
    ThLine2D::const_point_iterator pit = (*lit)->pointBegin();
    ThLine2D::const_point_iterator pit_end = (*lit)->pointEnd();
    for ( ; pit != pit_end; ++ pit ) {
      double x = (*pit)->x;
      double y = (*pit)->y;
      QGraphicsEllipseItem * ell = new QGraphicsEllipseItem( x-5, y-5, 10, 10 );
      ell->setBrush( QBrush( Qt::yellow  ) );
      ell->setPen( icon->PenYellow() );
      ell->setZValue( ZVALUE_GREY );
      ell->show();
      // addItem( ell );
      group -> addToGroup( ell );
    }
  }
  std::vector< ThPoint2D * >::const_iterator pit = scrap->pointsBegin();
  std::vector< ThPoint2D * >::const_iterator pit_end = scrap->pointsEnd();
  double scale = status->getScale();
  for ( ; pit != pit_end; ++pit ) {
    double x = (*pit)->x * scale;
    double y = (*pit)->y * scale;
    QGraphicsEllipseItem * ell = new QGraphicsEllipseItem( x-5, y-5, 10, 10 );
    ell->setBrush( QBrush( Qt::yellow  ) );
    ell->setPen( icon->PenYellow() );
    ell->setZValue( ZVALUE_GREY );
    ell->show();
    // addItem( ell );
    group -> addToGroup( ell );
  }
  group -> show();
}

void 
PlotCanvasScene::greyIn( TherionScrap * scrap )
{
  // fprintf(stderr, "greying in %s\n", scrap->getScrapName() );  
  QGraphicsItemGroup * group = scrap -> getGrey();
  if ( group ) {
    this -> removeItem( group );
    delete group;
    scrap -> setGrey( NULL );
  }
}

void
PlotCanvasScene::createPoint( ThPoint2D * pt )
{
#ifdef PLOT_PAINTER
  QGraphicsItem * item = plot_painter.createPoint( pt, status->getScale() );
  this -> addItem( item );
  g_points.push_back( std::pair< QGraphicsItem*, ThPoint2D*>( item, pt ) );
#else
  double x = pt->x;
  double y = pt->y;
  Therion::PointType type = pt->type();
  int orient = pt->orientation() * ORIENTATION_UNITS;

  double s = status->getScale();
  #ifdef HAS_THP_ICONS
    int thp_off = -12;
    QGraphicsPixmapItem * item =
      new QGraphicsPixmapItem( icon->ThpPixmap( type ) );
    item -> setOffset( thp_off, thp_off );
  #else
    // const QPolygon & poly = icon->ThpSymbol1( type );
    // QGraphicsPolygonItem * item = new QGraphicsPolygonItem( poly );
    const QPainterPath & path = icon->ThpSymbol2( type );
    QGraphicsPathItem * item = new QGraphicsPathItem( path );
    item->setBrush( icon->ThpBrush( type ) );
    item->setPen( icon->ThpPen( type ) );
  #endif
  
  switch ( type ) {
    case Therion::THP_AIR_DRAUGHT:
    case Therion::THP_ENTRANCE:
    case Therion::THP_WATER_FLOW:
      item -> rotate( orient );
      break;
    default:
      break;
  }
  pt->item = item;
  item -> setPos( x*s, y*s );
  item -> setZValue( ZVALUE_POINT );
  item -> show();
  this -> addItem( item );
  g_points.push_back( std::pair< QGraphicsItem*, ThPoint2D*>( item, pt ) );
#endif
}

void
PlotCanvasScene::createLabelPoint( ThPoint2D * pt, const char * text, int orient )
{
#ifdef PLOT_PAINTER
  QGraphicsItem * item = plot_painter.createLabelPoint( pt, status->getScale(),
                                                        text, orient );
  this -> addItem( item );
  g_points.push_back( std::pair<QGraphicsItem*, ThPoint2D*>( item, pt ) );
#else
  // fprintf(stderr, "createLabelPoint %s\n", text );
  double x = pt->x;
  double y = pt->y;
  double s = status->getScale();
  QString t("ABC");
  // if ( option ) {
  //   const char * ch1 = strchr( option, '"' );
  //   const char * ch2 = strrchr( option, '"' );
  //   if ( ch1 && ch2 && ch1 != ch2 ) {
  //     t = ch1+1;
  //     t.truncate( ch2 - ch1 - 1 );
  //   }
  // }
  if ( text && strlen(text) > 0 ) t = text;
  QGraphicsSimpleTextItem * txt = this ->addSimpleText( t );
  txt->rotate( orient * ORIENTATION_UNITS );
  txt->setPos( x*s, y*s );
  txt->setBrush( Qt::black );
  txt->show();
  g_points.push_back( std::pair<QGraphicsItem*, ThPoint2D*>( txt, pt ) );
  pt->item = txt;
#endif
}

// ---------------------------------------------------------------------

void
PlotCanvasScene::clearGrid()
{
  for (int k=0; k<grid_size; ++k ) {
    if ( k%10 != 5 ) {
      ig_grid->removeFromGroup( grid[k] );
    } else {
      ig_grid10->removeFromGroup( grid[k] );
    }
    this -> removeItem( grid[k] );
  }
  delete[] grid;
  grid = NULL;
}

void 
PlotCanvasScene::clearScaleBar()
{
  ig_scalebar -> removeFromGroup( scale_text );
  ig_scalebar -> removeFromGroup( scale_bar );
  this -> removeItem( scale_text );
  this -> removeItem( scale_bar );
}
  
void 
PlotCanvasScene::clearScrap()
{ 
  // clearStations();
  clearPoints();
  clearLines();
}

/** clear the drawing
 */
void 
PlotCanvasScene::clearPlot() 
{
  clearScrap();
  clearSurvey();
  #ifdef HAS_BACKIMAGE
    clearBackground();
  #endif
}

void 
PlotCanvasScene::clearSurvey()
{
  // fprintf(stderr, "clearSurvey()\n");
  clearStations();
  clearCenterline();
  clearSplay();
  #ifdef HAS_LRUD
    clearLrud();
  #endif
}  

#ifdef HAS_LRUD
void
PlotCanvasScene::clearLrud()
{
  // TODO
}
#endif

void
PlotCanvasScene::clearLines()
{
  DBG_CHECK("clearLines() size %d \n", line_segments.size() );
  std::vector< std::pair<QGraphicsLineItem *, ThLine2D *> >::iterator end = line_segments.end();
  std::vector< std::pair<QGraphicsLineItem *, ThLine2D *> >::iterator its = line_segments.begin();
  for ( ; its != end; ++its ) {
    // ig_lines -> removeFromGroup( its->first );
    this -> removeItem( its->first );
  }
  line_segments.clear();

  std::vector< std::pair<QGraphicsItem *, ThLine2D*> >::iterator endp = line_points.end();
  std::vector< std::pair<QGraphicsItem *, ThLine2D*> >::iterator itp  = line_points.begin();
  for ( ; itp != endp; ++itp ) {
    // ig_points -> removeFromGroup( itp->first );
    this -> removeItem( itp->first );
  }
  line_points.clear();
}

void
PlotCanvasScene::clearPoints()
{
  // fprintf(stderr, "clearPoints() size %d \n", g_points.size() );
  std::vector< std::pair<QGraphicsItem*, ThPoint2D*> >::iterator end = g_points.end();
  std::vector< std::pair<QGraphicsItem*, ThPoint2D*> >::iterator it  = g_points.begin();
  for ( ; it != end; ++it ) {
    // ig_points -> removeFromGroup( *it );
    this -> removeItem( it->first );
  }
  g_points.clear();
}

void 
PlotCanvasScene::clearCenterline()
{
  std::vector< std::pair<QGraphicsLineItem *,PlotSegment*> >::iterator end = centerline.end();
  std::vector< std::pair<QGraphicsLineItem *,PlotSegment*> >::iterator it  = centerline.begin();
  for ( ; it != end; ++it ) {
    ig_centerline -> removeFromGroup( it->first );
    this -> removeItem( it->first );
  }
  centerline.clear();
}


void 
PlotCanvasScene::clearSplay()
{
  std::vector< std::pair<QGraphicsLineItem *,PlotSegment*> >::iterator end = splay.end();
  std::vector< std::pair<QGraphicsLineItem *,PlotSegment*> >::iterator it  = splay.begin();
  for ( ; it != end; ++it ) {
    ig_splay -> removeFromGroup( it->first );
    this -> removeItem( it->first );
  }
  splay.clear();
}

void 
PlotCanvasScene::clearStations()
{
  DBG_CHECK("clearStations() size %d\n", stations.size() );
  std::vector< std::pair<QGraphicsSimpleTextItem *, PlotPoint * > >::iterator end = stations.end();
  std::vector< std::pair<QGraphicsSimpleTextItem *, PlotPoint * > >::iterator it  = stations.begin();
  for ( ; it != end; ++it ) {
    ig_stations -> removeFromGroup( it->first );
    this -> removeItem( it->first );
  }
  stations.clear();
}


#ifdef HAS_BACKIMAGE
// ---------------------------------------------------------------------
// BACKGROUND

void
PlotCanvasScene::clearBackground()
{
  /*
  std::vector< QGraphicsPixmapItem * >::iterator end = backgrounds.end();
  std::vector< QGraphicsPixmapItem * >::iterator it  = backgrounds.begin();
  for ( ; it != end; ++it ) {
    ig_background -> removeFromGroup( *it );
    this -> removeItem( *it );
  }
  backgrounds.clear();
  */
  if ( sketch_back ) {
    this -> removeItem( sketch_back );
    delete sketch_back;
  }
}

void 
PlotCanvasScene::evalBackground( const std::vector< BackgroundImageStation * > & stations,
                            QPixmap * back_pixmap )
{
  ARG_CHECK( back_pixmap == NULL, );
  DBG_CHECK("PlotCanvas::evalBackground() stations %d \n", stations.size() );

  if ( stations.size() < 2 ) { // caller should invoke this only 
    DBG_CHECK("PlotCanvas::evalBackground() too few stations %d \n", stations.size() );
    QMessageBox::warning(parent, lexicon("qtopo_plot"), lexicon("one_station") );
    return;                    // when stations are at least 2
  }
  backStations.clear();

  for ( std::vector< BackgroundImageStation * >::const_iterator it = stations.begin();
        it != stations.end();
        ++it ) {
    PlotPoint * pt = parent->getPlot()->getPoints();
    while ( pt && ( pt->name != (*it)->name ) ) pt=pt->next;
    if ( pt ) {
      DBG_CHECK("insert backStation at %.2f %.2f <--> %d %d\n", (*it)->x, (*it)->y, pt->x0, pt->y0 );
      backStations.push_back( BackgroundImagePoint( pt, *it ) );
      // it->x0 = 4 * (status->getOffsetX() + pt->x0);
      // it->y0 = 4 * (status->getOffsetY() + pt->y0);
    }
  }
  if ( backStations.size() < 2 ) {
    QMessageBox::warning(parent, lexicon("qtopo_plot"), lexicon("one_station") );
    return;
  }
  setBackgroundInit( back_pixmap );
}

void
PlotCanvasScene::setBackgroundInit( QPixmap * backPixmap )
{
  if ( backPixmap == NULL ) 
    return;

  // int w1 = (int)(width() * status->getScale() );
  // int h1 = (int)(height() * status->getScale() );
  int w1 = (int)( width() );
  int h1 = (int)( height() );
  int w2 = backPixmap->width();
  int h2 = backPixmap->height();
  DBG_CHECK("make background %d %d for pixmap %d %d \n", w1, h1, w2, h2 );
  // DBG_CHECK("make background %d %d (%.2f %.2f) for pixmap %d %d \n",
  //   w1, h1, width(), height(), w2, h2 );
  // DBG_CHECK("status offset %d %d \n", status->getOffsetX(), status->getOffsetY() );
  QImage src_image = backPixmap->toImage();
  const unsigned char * src = src_image.bits();

  QImage sketch( w1, h1, QImage::Format_RGB32 );
  unsigned char * dst = sketch.bits();

  memset( dst, 0xcc, w1*h1*4 ); // make background white

#ifdef WARP_PLAQUETTE
  therion::warp::morph_type morph_type = therion::warp::THMORPH_STATION;
  therion::warp::plaquette_algo algo;
#endif

  for ( std::vector< BackgroundImagePoint >::iterator it=backStations.begin();
        it != backStations.end();
        ++it ) {
    it->x0 = 4 * (status->getOffsetX() + it->point->x0);
    it->y0 = 4 * (status->getOffsetY() + it->point->y0);
    // DBG_CHECK("stations %.2f %.2f --> %.2f %.2f \n", it->x, it->y, it->x0, it->y0 );
#ifdef WARP_PLAQUETTE
    std::string name( it->point->name );
    algo.insert_point( morph_type, name, thvec2(it->x,it->y), thvec2(it->x0,it->y0) );
#endif
  }

#ifdef WARP_PLAQUETTE
  for ( std::vector< BackgroundImagePoint >::iterator it=backStations.begin();
        it != backStations.end();
        ++it ) {
    std::string name( it->point->name );
    std::vector< BackgroundImagePoint >::iterator it2 = it;
    for ( ++it2; it2 != backStations.end(); ++it2 ) {
      std::string name2( it2->point->name );
      if ( parent->GetList()->hasBlock( it->point->name, it2->point->name ) ) {
        algo.insert_line( morph_type, name, name2 );
        break;
      } 
    }
  }

  algo.initialize( therion::warp::THWARP_PLAN );
  thvec2 origin( w1/2.0, h1/2.0 );
  double unit = algo.to_unit() * 1.0;
  algo.map_image( src, w2, h2, dst, w1, h1, origin, unit, 4, therion::warp::THWARP_PLAN );
  if ( true ) {
#else
  DBG_CHECK("warp src %d %d --> dst %d %d \n", w2, h2, w1, h1 );
  if ( imageWarp( backStations, dst, w1, h1, src, w2, h2 ) ) {
#endif
    setBackground( & sketch );
  } else {
    DBG_CHECK("TODO tell the user that warping failed\n");
  }
}


void
PlotCanvasScene::setBackground( QImage * sketch )
{
  #ifdef SAVE_SKETCH
    if ( sketch->save("/tmp/sketch.png") ) {
      DBG_CHECK("saved sketch to /tmp/sketch.png %d %d \n", sketch->width(), sketch->height() );
    } else {
      DBG_CHECK("failed to save sketch\n");
    }
  #endif
  QSize size( (int)(sketch->width() * (status->getScale())/4 ),
              (int)(sketch->height() * (status->getScale())/4 ) );
  DBG_CHECK("PlotCanvasScene::setBackground() %d %d \n", size.width(), size.height());
  QImage img2 = sketch->scaled( size, Qt::KeepAspectRatio );
  QPixmap sketch_pix = QPixmap::fromImage( img2 );
  if ( sketch_back ) {
    this -> removeItem( sketch_back );
    delete sketch_back;
  }
  sketch_back = this -> addPixmap( sketch_pix );
  sketch_back->setVisible( true );
  sketch_back->setZValue( ZVALUE_BACK );
  // ig_background -> addToGroup( sketch_back );
  // backgrounds.push_back( sketch_back );
}

#endif // HAS_BACKIMAGE

// ===========================================================================
// LABEL WIDGET
// "label" and "continuation" points text

LabelWidget::LabelWidget( PlotCanvasScene * p, const char * caption )
  : QDialog( NULL /* p->patent */  )
  , parent( p )
{
  setWindowTitle( caption );
  QVBoxLayout* vbl = new QVBoxLayout(this);
  text = new QLineEdit( "", this );
  connect( text, SIGNAL(returnPressed()), this, SLOT(doOK()) );
  vbl->addWidget( text );
  // show();
  exec();
}
  
// =======================================================================
// POINT OPTIONS
PointOptionsWidget::PointOptionsWidget( PlotCanvasScene * scene, ThPoint2D * pt )
  : parent( scene )
  , point( pt )
  , options( NULL )
  , subtype( NULL )
  , align( NULL )
  , scale( NULL )
  , porient( NULL )
{
  Language & lexicon = Language::Get();
  setWindowTitle( lexicon( "point_options" ) );
  QVBoxLayout* vbl = new QVBoxLayout(this);

  int t = (int)( point->type() );

  QWidget * hb;
  QHBoxLayout * hbl;
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  QLabel * label = new QLabel( "type", this );
  hbl->addWidget( label );
  QLabel * type = new QLabel( Therion::PointName[ t ], this );
  hbl->addWidget( type );
  vbl->addWidget( hb );

  const char ** st = &(Therion::PointSubtype[ t ][0]);
  if ( *st != NULL ) {
    hb = new QWidget(this);
    hbl = new QHBoxLayout( hb );
    label = new QLabel( "subtype", this );
    hbl->addWidget( label );
    subtype = new QComboBox( this );
    int n = 0;
    while ( *st ) {
      subtype->addItem( *st );
      if ( *st == point->subtype() ) subtype->setCurrentIndex( n );
      ++ st;
      ++ n;
    }
    subtype->addItem( " " );
    if ( *st == point->subtype() ) subtype->setCurrentIndex( n );
    hbl->addWidget( subtype );
    vbl->addWidget( hb );
  }

  st =  &(Therion::PointAlign[0]);
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  label = new QLabel( "align", this );
  hbl->addWidget( label );
  align = new QComboBox( this );
  int n = 0;
  while ( *st ) {
    align->addItem( *st );
    if ( *st == point->align() ) align->setCurrentIndex( n );
    ++ st;
    ++ n;
  }
  align->addItem( " " );
  if ( *st == point->align() ) align->setCurrentIndex( n );
  hbl->addWidget( align );
  vbl->addWidget( hb );

  st =  &(Therion::PointScale[0]);
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  label = new QLabel( "scale", this );
  hbl->addWidget( label );
  scale = new QComboBox( this );
  n = 0;
  while ( *st ) {
    scale->addItem( *st );
    if ( *st == point->scale() ) scale->setCurrentIndex( n );
    ++ st;
    ++ n;
  }
  scale->addItem( " " );
  if ( *st == point->scale() ) scale->setCurrentIndex( n );
  hbl->addWidget( scale );
  vbl->addWidget( hb );

  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  label = new QLabel( "options", this );
  hbl->addWidget( label );
  options = new QLineEdit( point->option(), this );
  hbl->addWidget( options );
  vbl->addWidget( hb );
 
  if ( Therion::PointHasOrientation[t] ) {
    hb = new QWidget(this);
    hbl = new QHBoxLayout( hb );
    hbl->addWidget( new QLabel( lexicon("orientation"), hb ) );
    porient = new QDial( hb );
    porient->setMinimum( 0 );
    porient->setMaximum( 360 );
    porient->setWrapping( true );
    porient->setNotchesVisible( true );
    porient->setNotchTarget( ORIENTATION_UNITS );
    porient->resize( 40, 40 );
    // DBG_CHECK("orientation %d \n", scene->getPointOrient() * ORIENTATION_UNITS - 180 );
    porient->setValue( (point->orientation() * ORIENTATION_UNITS + 180)%360 );
    hbl->addWidget( porient );
    vbl->addWidget( hb );
  }

  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  QPushButton * c = new QPushButton( tr(lexicon("ok")), hb );
  connect( c, SIGNAL(clicked()), this, SLOT(doOK()) );
  hbl->addWidget( c );
  c = new QPushButton( tr(lexicon("cancel")), hb);
  connect( c, SIGNAL(clicked()), this, SLOT(doCancel()) );
  hbl->addWidget( c );
  vbl->addWidget( hb );

  exec();
}

void
PointOptionsWidget::doOK()
{
  hide();
  if ( subtype ) {
    int j = subtype->currentIndex();
    point->setSubtype( Therion::PointSubtype[ point->type() ][ j ] );
  }
  if ( align ) {
    int j = align->currentIndex();
    if ( j == -1 ) j = THP_ALIGN_MAX - 1;
    assert ( j >= 0 && j < THP_ALIGN_MAX );
    point->setAlign( Therion::PointAlign[ j ] );
  }
  if ( scale ) {
    int j = scale->currentIndex();
    if ( j == -1 ) j = THP_SCALE_MAX-1;
    assert ( j >= 0 && j < THP_SCALE_MAX );
    point->setScale( Therion::PointScale[ j ] );
  }
  if ( ! options->text().isEmpty() ) {
    point->setOption( options->text().TO_CHAR() );
  }
  if ( porient ) {
    int old_angle = point -> orientation();
    int angle = ( ( porient->value() + 180 + ORIENTATION_HALF) / ORIENTATION_UNITS ) % (ORIENTATION_MAX);
    point -> setOrientation( angle );
    point -> getItem() -> rotate( (angle - old_angle)*ORIENTATION_UNITS );
  }
}

// =======================================================================
// AREA POINT OPTIONS
AreaPointOptionsWidget::AreaPointOptionsWidget( PlotCanvasScene * scene, 
                                                ThLinePoint * pt, 
                                                ThArea2D * ar )
  : parent( scene )
  , point( pt )
  , area( ar )
  , visible( NULL )
{
  Language & lexicon = Language::Get();
  setWindowTitle( lexicon( "areapoint_options" ) );
  QVBoxLayout* vbl = new QVBoxLayout(this);

  int t = (int)( area->Type() );

  QWidget * hb;
  QHBoxLayout * hbl;
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  QLabel * label = new QLabel( "area type", this );
  hbl->addWidget( label );
  QLabel * type = new QLabel( Therion::AreaName[ t ], this );
  hbl->addWidget( type );
  vbl->addWidget( hb );

  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  visible = new QCheckBox( tr(lexicon("border_visible")), this );
  hbl->addWidget( visible );
  vbl->addWidget( hb );
  visible->setChecked( area->isBorderVisible() );

  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  QPushButton * c = new QPushButton( tr(lexicon("ok")), hb );
  connect( c, SIGNAL(clicked()), this, SLOT(doOK()) );
  hbl->addWidget( c );
  c = new QPushButton( tr(lexicon("cancel")), hb);
  connect( c, SIGNAL(clicked()), this, SLOT(doCancel()) );
  hbl->addWidget( c );
  vbl->addWidget( hb );

  exec();
}

void
AreaPointOptionsWidget::doOK()
{
  hide();
  if ( visible ) {
    area->setBorderVisible( visible->isChecked() );
  }
}
  
// =======================================================================
// LINE POINT OPTIONS
LinePointOptionsWidget::LinePointOptionsWidget( PlotCanvasScene * scene, 
                                                ThLinePoint * pt, ThLine2D * ln )
  : parent( scene )
  , point( pt )
  , line( ln )
  , options( NULL )
  , subtype( NULL )
  , lsize( NULL )
  , clipped( NULL )
  , reversed( NULL )
  , visible( NULL )
  , special_cb( NULL )
  , special_le( NULL )
{
  Language & lexicon = Language::Get();
  setWindowTitle( lexicon( "linepoint_options" ) );
  QVBoxLayout* vbl = new QVBoxLayout(this);

  int t = (int)( line->type() );

  QWidget * hb;
  QHBoxLayout * hbl;
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  QLabel * label = new QLabel( "line type", this );
  hbl->addWidget( label );
  QLabel * type = new QLabel( Therion::LineName[ t ], this );
  hbl->addWidget( type );
  vbl->addWidget( hb );

  const char ** st = &(Therion::LineSubtype[ t ][0]);
  if ( *st != NULL ) {
    hb = new QWidget(this);
    hbl = new QHBoxLayout( hb );
    label = new QLabel( "line subtype", this );
    hbl->addWidget( label );
    subtype = new QComboBox( this );
    int n = 0;
    while ( *st ) {
      subtype->addItem( *st );
      if ( *st == line->getSubtype() ) subtype->setCurrentIndex( n );
      ++ st;
      ++ n;
    }
    subtype->addItem( " " );
    if ( *st == line->getSubtype() ) subtype->setCurrentIndex( n );
    hbl->addWidget( subtype );
    vbl->addWidget( hb );
  }

  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  label = new QLabel( "outline", this );
  hbl->addWidget( label );
  outline_cb = new QComboBox( this );
  int n = 0;
  const char ** ot = &(Therion::LineOutline[0]);
  while ( *ot ) {
    outline_cb -> addItem( *ot );
    if ( *ot == line->getOutline() ) outline_cb -> setCurrentIndex( n );
    ++ ot;
    ++ n;
  }
  outline_cb -> addItem( " " );
  if ( *ot == line->getOutline() ) outline_cb -> setCurrentIndex( n );
  hbl->addWidget( outline_cb );
  vbl->addWidget( hb );

  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  reversed = new QCheckBox( "reversed", this );
  reversed->setChecked( line->isReversed() );
  hbl->addWidget( reversed );

  
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  clipped = new QCheckBox( "clipped", this );
  clipped->setChecked( line->isClipped() );
  hbl->addWidget( clipped );
  vbl->addWidget( hb );
  
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  visible = new QCheckBox( "visible", this );
  visible->setChecked( line->isVisible() );
  hbl->addWidget( visible );
  vbl->addWidget( hb );

  switch ( line->type() ) {
    case Therion::THL_ARROW:
    case Therion::THL_SLOPE:
    {
      hb = new QWidget(this);
      hbl = new QHBoxLayout( hb );
      label = new QLabel( Therion::LineSpecialOptionName[t], this );
      hbl->addWidget( label );
      special_cb = new QComboBox( this );
      int n = 0;
      const char ** sp = &(Therion::LineSpecialOption[t][0]);
      while ( *sp ) {
        special_cb -> addItem( *sp );
        if ( *st == line->getSpecialString() ) special_cb -> setCurrentIndex( n );
        ++ sp;
        ++ n;
      }
      special_cb -> addItem( " " );
      if ( *sp == line->getSpecialString() ) special_cb -> setCurrentIndex( n );
      hbl->addWidget( special_cb );
      vbl->addWidget( hb );
      break;
    }
    case Therion::THL_PIT:
    case Therion::THL_WALL:
    {
      hb = new QWidget(this);
      hbl = new QHBoxLayout( hb );
      label = new QLabel( Therion::LineSpecialOptionName[t], this );
      hbl->addWidget( label );
      QString value(" ");
      if ( line->hasSpecialValue() ) value.setNum( line->getSpecialValue() );
      special_le = new QLineEdit( value, this );
      hbl->addWidget( special_le );
      vbl->addWidget( hb );
      break;
    }
    default:
      break;
  }
  
  if ( t == Therion::THL_SLOPE ) { // t = Therion::THL_SLOPE || t == Therion::THL_GRADIENT
    hb = new QWidget(this);
    hbl = new QHBoxLayout( hb );
    label = new QLabel( "point l-size", this );
    hbl->addWidget( label );
    if ( point->getLSize() > 0 ) {
      QString t;
      t.setNum( point->getLSize() );
      lsize = new QLineEdit( t, this );
    } else {
      lsize = new QLineEdit( "", this );
    }
    hbl->addWidget( lsize );
    vbl->addWidget( hb );
  }

  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  label = new QLabel( "point adjust", this );
  hbl->addWidget( label );
  const char ** ad = &(Therion::LinePointAdjust[0]);
  adjust = new QComboBox( this );
  n = 0;
  while ( *ad ) {
    adjust->addItem( *ad );
    if ( point->getAdjust() == *ad ) adjust->setCurrentIndex( n );
    ++ ad;
    ++ n;
  }
  adjust->addItem( " " );
  if ( *ad == point->getAdjust() ) adjust->setCurrentIndex( n );
  hbl->addWidget( adjust );
  vbl->addWidget( hb );
  
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  label = new QLabel( "point options", this );
  hbl->addWidget( label );
  options = new QLineEdit( point->LPoption(), this );
  hbl->addWidget( options );
  vbl->addWidget( hb );
 
  hb = new QWidget(this);
  hbl = new QHBoxLayout( hb );
  QPushButton * c = new QPushButton( tr(lexicon("ok")), hb );
  connect( c, SIGNAL(clicked()), this, SLOT(doOK()) );
  hbl->addWidget( c );
  c = new QPushButton( tr(lexicon("cancel")), hb);
  connect( c, SIGNAL(clicked()), this, SLOT(doCancel()) );
  hbl->addWidget( c );
  vbl->addWidget( hb );

  exec();
}

void
LinePointOptionsWidget::doOK()
{
  hide();
  if ( subtype ) {
    int j = subtype->currentIndex();
    line->setSubtype( Therion::LineSubtype[ line->type() ][ j ] );
  }

  line -> setOutline( Therion::LineOutline[ outline_cb -> currentIndex() ] );

  if ( reversed ) {
    line->setReversed( reversed->isChecked() );
  }
  if ( visible ) {
    line->setVisible( visible->isChecked() );
  }
  if ( clipped ) {
    line->setClipped( clipped->isChecked() );
  }
  if ( special_cb ) {
    line->setSpecialString( 
      Therion::LineSpecialOption[line->type()][ special_cb -> currentIndex() ] );
  } else if ( special_le && ! special_le->text().isEmpty() ) {
    line->setSpecialValue( special_le->text().toInt() );
  }

  if ( lsize && ! lsize->text().isEmpty() ) {
    int s = lsize->text().toInt();
    if ( s >= 0 ) {
      point->setLSize( s );
    }
  }
  if ( options && ! options->text().isEmpty() ) {
    point->setLPOption( options->text().TO_CHAR() );
  }
}
