/** @file PlotCanvasScene.h
 * 
 * @author marco corvi
 * @date dec 2010
 * 
 * @brief Scene of a plot
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#ifndef PLOT_CANVAS_SCENE_H
#define PLOT_CANVAS_SCENE_H
#include <assert.h>

#include <vector>
#include <utility>

#include <QGraphicsScene>
#include <QGraphicsItem>
#include <QGraphicsLineItem>
#include <QGraphicsPathItem>
#include <QGraphicsSimpleTextItem>
#include <QGraphicsPolygonItem>
#include <QGraphicsItemGroup>
#include <QDialog>
#include <QLineEdit>
#include <QPainterPath>
#include <QComboBox>
#include <QDial>

#include "shorthands.h"
#include "IconSet.h"
#include "Language.h"
#ifdef HAS_BACKIMAGE
  #include "BackgroundImage.h"
  #include "ImageTransform.h"
#endif
#include "PlotSegment.h"
#include "PlotPoint.h"
#include "TherionPoint.h"
#include "TherionLine.h"
#include "Units.h"
#include "CanvasMode.h"
#include "PlotStatus.h"
#include "PlotLineItem.h"

#include "PlotPainter.h"
#include "PlotAreaItem.h"

#define ZOOM_10 (-5)

class PlotCanvas; // forward


class PlotCanvasScene : public QGraphicsScene
                 #ifdef HAS_BACKIMAGE
                 , public BackgroundImageCallback
                 #endif
{
  private:
    PlotCanvas * parent;
    IconSet * icon;             //!< pointer to the singleton icon set
    Language & lexicon;         //!< i18n lexicon
    PlotPainter plot_painter;   //!< graphics item creator

    std::vector< std::pair<QGraphicsSimpleTextItem *, PlotPoint*> > stations;
    std::vector< std::pair<QGraphicsLineItem *, PlotSegment*> > centerline;
    std::vector< std::pair<QGraphicsLineItem *, PlotSegment*> > splay;
    #ifdef HAS_LRUD
      std::vector< QGraphicsLineItem * > lrud;
    #endif
    std::vector< std::pair<QGraphicsItem *, ThPoint2D *> > g_points;
    std::vector< std::pair<QGraphicsItem *, ThLine2D *> > line_points;
    std::vector< std::pair<PlotLineItem *, ThLine2D *> > line_segments;
    std::vector< std::pair<PlotAreaItem *, ThArea2D *> > areas;
    std::vector< std::pair<QGraphicsItem *, ThArea2D *> > area_points;
    std::vector< std::pair<PlotLineItem *, ThArea2D *> > area_segments;
    // std::vector< std::pair<QGraphicsItem *, ThLinePoint *> > line_points;
    // std::vector< std::pair<QGraphicsLineItem *, ThLine2D *> > lines;


    #ifdef HAS_BACKIMAGE
      // std::vector< QGraphicsPixmapItem * > backgrounds;
    #endif

    QGraphicsItemGroup * ig_stations;   
    QGraphicsItemGroup * ig_centerline; // this is always shown
    QGraphicsItemGroup * ig_splay;
    #ifdef HAS_LRUD
      QGraphicsItemGroup * ig_lrud;
    #endif
    QGraphicsItemGroup * ig_grid;        //!< group of grid lines
    QGraphicsItemGroup * ig_grid10;      //!< group of grid 10 lines
    QGraphicsItemGroup * ig_scalebar;    //!< group of scalebar items
    // QGraphicsItemGroup * ig_points;   // this is not really necessary as it is always shown
    // QGraphicsItemGroup * ig_lines;    // same
    // QGraphicsItemGroup * ig_areas;    // same

    bool on_line;             //!< whether it is drawing a line
    bool on_area;             //!< whether it is drawing an area
    bool do_splay;            //!< whether to show splay shots

    const Units & units;      //!< reference to the current units
    PlotStatus * status;      //!< non-volatile status of the plot (set by the PlotCanvas)
    int m_zoom;               //!< zoom value (default 0, pos. zoom in, neg. zoom out )

    enum InputMode input_mode;       //!< current canvas input mode
    Therion::PointType point_type;   //!< current point type
    int         point_orient;        //!< current point orientation
    std::string label_text;          //!< text of the current label point

    Therion::LineType  line_type;    //!< current line type
    ThLine2D * cur_line;             //!< current (therion) line 
    QGraphicsItem * temp_pt;         //!< temporary circle to mark mouse press 
    QGraphicsItem * temp_cp;         //!< temporary circle to mark mouse press 

    Therion::AreaType area_type;
    ThArea2D * cur_area;

    double cur_pt_x, cur_pt_y;       //!< X/Y coordinates of mouse press

    ThPoint2D * moving_pt;
    ThLine2D *  moving_line;
    ThArea2D * moving_area;
    ThLinePoint * moving_line_pt;
    QGraphicsEllipseItem * moving_item;

    QGraphicsLineItem ** grid;            //!< plot grid
    int grid_size;                        //!< number of lines of the grid
    QGraphicsLineItem * scale_bar;        //!< scale bar (10 m)
    QGraphicsSimpleTextItem * scale_text; //!< scale value 

  #ifdef HAS_BACKIMAGE        // background sketch
    // QGraphicsItemGroup * ig_background;
    QGraphicsPixmapItem * sketch_back;
    std::vector< BackgroundImagePoint > backStations;
  #endif


  public:
    PlotCanvasScene( int x, int y, int w, int h, PlotCanvas * p );

    ~PlotCanvasScene();

    /** grey out the items of a group
     */
    void greyOut( TherionScrap * scrap );

    /** grey in (make them foreground) the items of a group
     */
    void greyIn( TherionScrap * scrap );


    /** set the (temporary) text of the label
     * @param text  label text
     */
    void setLabelText( const char * text ) { label_text = text; }

    /** change zoom
     * @param dz  value by which the zoom change
     */
    void zoom( int dz ) 
    { 
      m_zoom += dz; 
      if ( status->getStatus() == PLOT_FRAME_GRID ) {
        if ( m_zoom > ZOOM_10 ) {
          ig_grid->show();
        } else {
          ig_grid->hide();
        }
        // ig_grid10->show();
      }
    }

    /** set whether is drawing a line or not
     * @param yes_no
     */
    void setOnLine( bool yes_no );
    bool onLine() const { return on_line; }

    /** set whether is drawing an area or not
     * @param yes_no 
     */
    void setOnArea( bool yes_no );
    bool onArea() const { return on_area; }

    /** set the current therion line
     * @param line   new current therion line
     */
    void setCurLine( ThLine2D * line ) { cur_line = line; }
    ThLine2D * curLine() { return cur_line; }

    /** set the current therion area
     * @param area   new current therion area
     */
    void setCurArea( ThArea2D * area ) { cur_area = area; }
    ThArea2D * curArea() { return cur_area; }

    /** accessor: get the input mode
     * @return the current input mode
     */
    InputMode getInputMode() const { return input_mode; }
    void setInputMode( InputMode mode );

    /** accessor to the point type
     * @return the current point type
     */
    Therion::PointType getPointType() const { return point_type; }
    void setPointType(int t ) { point_type = (Therion::PointType)t; }

    /** set the point orientation
     * @param orient   point orientation
     */
    void setPointOrient( int orient ) { point_orient = orient; }
    int getPointOrient() const { return point_orient; }


    void setOffset();

    void dropPoint( ThPoint2D * pt );

    void dropLinePoint();
    void dropLine();

    void dropAreaPoint();
    void dropArea();

    void uncloseLine( ThLine2D * line );
    void unfinishArea( ThArea2D * area );

    /** set the type of line
     * @param t      new line type
     * @param is_line_mode whether it is in line mode
     */
    void setLineType(int t, bool is_line_mode );
    Therion::LineType getLineType() const { return line_type; }

    void setAreaType(int t, bool is_area_mode );
    Therion::AreaType getAreaType() const { return area_type; }

    /** reset line type (called by the UNDO)
     * @param t      new line type
     */
    void resetLineType( int t ) { line_type = (Therion::LineType)t; }
    void resetAreaType( int t ) { area_type = (Therion::AreaType)t; }

    // PlotStatus * getStatus() { return status; }
    // 1void setStatus( PlotStatus * s ) { status = s; }

    void mousePressEvent( QGraphicsSceneMouseEvent * e0 );
    void mouseReleaseEvent( QGraphicsSceneMouseEvent * e0 );

    void showPlot( );

    void toggleSplay() {
      do_splay = ! do_splay;
      if ( do_splay ) {
        ig_splay->show();
      } else {
        ig_splay->hide();
      }
    }
    
    void displayNumbers( );
    void displayGrid( int s );

    void makeGrid( int d0, int gx, int gy, int dx, int dy );

    void makeScaleBar( int x1, int x2, int y, const char * scale );

    void makeSurvey();
    void clearSurvey();

  public:
    void clearScrap();

    /** clear the drawing
     */
    void clearPlot();



  private:
    void clearGrid();
    void clearScaleBar();
    void clearCenterline();
    void clearSplay();
    void clearStations();
    void clearPoints();
    void clearLines();
    #ifdef HAS_LRUD
      void clearLrud();
    #endif
    #ifdef HAS_BACKIMAGE
      void clearBackground();
    #endif

    void rotatePenBrush( QPen & pen, double x1, double y1, double x2, double y2 );

    /** create the graphics polygon of an area
     * @param area   area
     * @return the graphics polygon
     */
    // QGraphicsPolygonItem * createAreaPolygon( ThArea2D * area );

  public:
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    // LINE

    /** insert a line control point (PlotDrawer interface)
     * @param x       X coordinate
     * @param y       Y coordinate
     * @param type    line type (used only when a new line begins)
     * @param add_offset whether to add the offset (default no)
     * @return the last inserted point (or NULL if the line is closed)
     *
     * @note used by QTshot
     */
    ThLinePoint * insertLinePoint( double x, double y, double x1, double y1,
                                   Therion::LineType type, bool add_offset = false );

    ThLinePoint * insertAreaPoint( double x, double y, double x1, double y1, 
                                   Therion::AreaType type, bool add_offset = false );

  private:
    /** set the control points for a line point
     * @param x   X coord of the second control point
     * @param y   Y coord of the second control point
     * @param pt  line point
     * @param add_offset whether to add the offset (default no)
     */
    void setControlPoints( double x2, double y2, ThLinePoint * pt, bool add_offset = false );
  public:
    void setControlPoint1( double x1, double y1, ThLinePoint * pt, bool add_offset = false );
    void setControlPoint2( double x2, double y2, ThLinePoint * pt, bool add_offset = false );

  public:
    /** close the current line if it is open
     * @param do_assert whether to assert
     * @note used by PlotCanvas dstr
     */
    void endLine( bool do_assert = true );

    /** finish current open area
     */
    void finishArea();

    void endArea( bool do_assert = true );

  public: // FIXME only for PlotTh2Import
    /**
     * @param type   line type
     * @param size   line size before inserting the point
     * @return pointer to the graphics line point item 
     */
    void createLineSegmentTo( ThLinePoint * line_pt, Therion::LineType type, size_t size,
                          QGraphicsItem ** ppt, PlotLineItem ** psgm );

  private:
    void createAreaSegmentTo( ThLinePoint * area_pt, Therion::AreaType type, size_t size,
                              QGraphicsItem ** ppt, PlotLineItem ** psgm );


    /** get the segment associated with a Canvas line
     * @param line   canvas line
     * @return the associated segment (or NULL)
     */
    PlotSegment * getSegmentFromLine( QGraphicsLineItem * line );

    PlotSegment * getSegmentFromArea( QGraphicsLineItem * area );

  public:
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    // POINT

    /** insert a point on the list of therion 2D points
     * _param x       X coordinate
     * @param y       Y coordinate
     * @param type    point type
     * @param orient  point orientation (WATER and AIR only)
     * @param option  point option string
     * @param add_offset whether to add the offset (default no)
     */
    ThPoint2D * insertPoint( double x, double y, 
                      Therion::PointType type, int orient = 0, 
                      const char * option = NULL,
                      bool add_offset = false );

    ThPoint2D * insertPointLabel( double x, double y,
                           Therion::PointType type, 
                           const char * text,
                           int orient = 0 );

  private:
    /** make a ThPoint2D
     * @param x  scene (or PT) X coord
     * @param y  scene (or PT) Y coord
     * @param type point type
     * @param add_offset whether to add the offset to the coords
     */
    ThPoint2D * makePoint( double x, double y, Therion::PointType type,
                           bool add_offset = false );

    /** create a display point with orientation
     */
    void createPoint( ThPoint2D * pt );

    /** create a display label point
     * @param text label/continuation display text
     * @param orient display orientation
     */
    void createLabelPoint( ThPoint2D * pt, const char * text, int orient);

    /** clear list of items (lines and points) of the plot
     * @param lines   vector of lines
     * @param pts     vector of points
     *
    void ClearTh2( std::vector< ThLine2D * > & lines, std::vector< ThPoint2D > & pts );
     */

    /** scan the list of text items and return the point for the given item
     * @param text
     * @return pointer to the canvas point
     * @note inefficient: O(N)
     */ 
    PlotPoint * getTextPoint( QGraphicsSimpleTextItem * text );
  
  public:
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // BACKGROUND IMAGE

    #ifdef HAS_BACKIMAGE
      /** make the background image warping the sketch
       * @param stations vector of station correspondences
       * @param back_pixmap sketch background pixmap
       */
      void evalBackground( const std::vector< BackgroundImageStation * > & stations,
                           QPixmap * back_pixmap );

      /** set the background image
       * @param sketch   background transformed (warped) image
       */
      void setBackground( QImage * sketch );

      /** initialize the background sketch image
       * @param backPixmap  background pixmap
       */
      void setBackgroundInit( QPixmap * backPixmap );
    #endif
};

// ------------------------------------------------------------
/** dialog to get the label text, input by the user
 */
class LabelWidget : public QDialog
{
  Q_OBJECT
  private:
    PlotCanvasScene * parent;
    QLineEdit * text;

  public:
    LabelWidget( PlotCanvasScene * scene, const char * caption );

  public slots:
    void doOK() 
    { 
      // fprintf(stderr, "Set label to parent \"%s\"\n", text->text().TO_CHAR() );
      parent->setLabelText( text->text().TO_CHAR() );
      done( 0 );
    }
};

class PointOptionsWidget : public QDialog
{
  Q_OBJECT
  private:
    PlotCanvasScene * parent; //!< scane parent
    ThPoint2D * point;        //!< therion point
    QLineEdit * options;      //!< point options
    QComboBox * subtype;
    QComboBox * align;
    QComboBox * scale;
    QDial * porient;          //!< point orientation dial
  

  public:
    PointOptionsWidget( PlotCanvasScene * scene, ThPoint2D * pt );

  public slots:
    void doOK();

    void doCancel() { hide(); }

};


class LinePointOptionsWidget : public QDialog
{
  Q_OBJECT
  private:
    PlotCanvasScene * parent; //!< scane parent
    ThLinePoint * point;      //!< therion linepoint
    ThLine2D * line;          //~< therion line
    QLineEdit * options;      //!< linepoint options
    QComboBox * subtype;      //!< line subtype
    QLineEdit * lsize;
    QCheckBox * clipped;
    QCheckBox * reversed;
    QCheckBox * visible;
    QComboBox * outline_cb;
    QComboBox * adjust;
    QComboBox * special_cb;
    QLineEdit * special_le;
  

  public:
    LinePointOptionsWidget( PlotCanvasScene * scene, ThLinePoint * pt, ThLine2D * ln );

  public slots:
    void doOK();

    void doCancel() { hide(); }

};

class AreaPointOptionsWidget : public QDialog
{
  Q_OBJECT
  private:
    PlotCanvasScene * parent; //!< scane parent
    ThLinePoint * point;      //!< therion linepoint
    ThArea2D * area;          //~< therion area
    QCheckBox * visible;
  

  public:
    AreaPointOptionsWidget( PlotCanvasScene * scene, ThLinePoint * pt, ThArea2D * ar );

  public slots:
    void doOK();

    void doCancel() { hide(); }

};


#endif
