/** @file PlotPainter.cpp
 *
 * @author marco corvi
 * @date feb 2011
 *
 * @brief plot therion area
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#include <QColor>
#include <QPen>
#include <QGraphicsItem>
#include <QGraphicsPolygonItem>
#include <QGraphicsEllipseItem>
#include <QGraphicsLineItem>
#include <QGraphicsPathItem>

#include "ArgCheck.h"
#include "PlotPainter.h"
#include "PlotZValue.h"
#include "TherionLine.h"

QGraphicsItem * 
PlotPainter::createLinePoint( ThLinePoint * line_pt,  ThLinePoint * prev, Therion::LineType type )
{
  double x = line_pt -> x;
  double y = line_pt -> y;
  QColor color = icon->LinePointColor( type );
  QGraphicsItem * pt = NULL;
  if ( type == Therion::THL_ARROW ) {
    if ( prev == NULL ) {
      pt = createRoundLinePoint( x, y, color );
    } else {
      int k = 0;
      double dx = x - prev->x;
      double dy = y - prev->y;
      if ( fabs(dx) > 2*fabs(dy) ) {
        k = (dx > 0)? 2 : 6;
      } else if ( fabs(dy) > 2*fabs(dx) ) { 
        k = (dy > 0)? 4 : 0;
      } else if ( dx > 0 ) {
        k = (dy > 0)? 3 : 1;
      } else {
        k = (dy > 0)? 5 : 7;
      }
      pt = createPolyPoint( x, y, icon->ArrowEnd(k) );
    }
  } else {
    pt = createRoundLinePoint( x, y, color );
  }
  return pt;
}

QGraphicsItem *
PlotPainter::createLineSegment( ThLinePoint * from, ThLinePoint * to, Therion::LineType type )
{
  QGraphicsLineItem * sgm = NULL;
  if ( from != NULL ) {
    QColor color;
    QPen pen;
    icon->LineStyle( type, pen, color );
    if ( type == Therion::THL_PIT || type == Therion::THL_CHIMNEY ) {
      rotatePenBrush( pen, from->x, from->y, to->x, to->y );
    }
    sgm = createLineSegment( from->x, from->y, to->x, to->y, pen );
  }
  return sgm;
}

void
PlotPainter::createLineSegmentTo( ThLinePoint * prev,
                                  ThLinePoint * cur,
                                  Therion::LineType type,
                                  QGraphicsItem ** ppt, QGraphicsLineItem ** psgm )
{
  double x = cur->x;
  double y = cur->y;
  if ( ppt ) {
    // create end-point
    QColor color = icon->LinePointColor( type );
    QGraphicsItem * pt = NULL;
    if ( type == Therion::THL_ARROW ) {
      if ( prev == NULL ) {
        pt = createRoundLinePoint( x, y, color );
      } else {
        int k = 0;
        double dx = x - prev->x;
        double dy = y - prev->y;
        if ( fabs(dx) > 2*fabs(dy) ) {
          k = (dx > 0)? 2 : 6;
        } else if ( fabs(dy) > 2*fabs(dx) ) { 
          k = (dy > 0)? 4 : 0;
        } else if ( dx > 0 ) {
          k = (dy > 0)? 3 : 1;
        } else {
          k = (dy > 0)? 5 : 7;
        }
        pt = createPolyPoint( x, y, icon->ArrowEnd(k) );
      }
    } else {
      pt = createRoundLinePoint( x, y, color );
    }
    *ppt = pt;
  }

  // FIXME use special brushes for pit and chimney
  if ( psgm ) {
    // create line-segment
    QGraphicsLineItem * sgm = NULL;
    if ( prev != NULL ) {
      QColor color;
      QPen pen;
      icon->LineStyle( type, pen, color );
      if ( type == Therion::THL_PIT || type == Therion::THL_CHIMNEY ) {
        rotatePenBrush( pen, prev->x, prev->y, x, y );
      }
      sgm = createLineSegment( prev->x, prev->y, x, y, pen );
#if 0
      if ( type == Therion::THL_PIT || type == Therion::THL_CHIMNEY ) {
        double dx = y - prev->y;
        double dy = x - prev->x;
        double d = ( dx*dx + dy*dy);
        if ( d > 0.0 ) {
          d = sqrt( d );
          if ( type == Therion::THL_CHIMNEY ) {
            d = -d; // ticks point into the cave
            pen = icon->PenViolet();
          }
          dx /= d;
          dy /= d;
          createLineSegment( x, y, x+PIT_TICK*dx, y-PIT_TICK*dy, pen );
        }
      }
#endif
    }
    *psgm = sgm;
  }
}

// used for LINE points
QGraphicsEllipseItem *
PlotPainter::createRoundLinePoint( double x, double y, const QColor & color )
{
  DBG_CHECK("createRoundLinePoint %.2f %.2f\n", x, y );
  QGraphicsEllipseItem * ell = new QGraphicsEllipseItem( x-2, y-2, 4, 4 );
  ell->setPen(  color );
  ell->setBrush( QBrush( color ) );
  ell->setZValue( ZVALUE_LINE_PT );
  ell->show();
  // this -> addItem( ell );
  // line_points.push_back( std::pair< QGraphicsItem*, ThLine2D*>(ell,cur_line) );
  return ell;
}

// used for ARROW heads
QGraphicsItem *
PlotPainter::createPolyPoint( double x, double y, const QPolygon & pts )
{
  QGraphicsPolygonItem * poly = new QGraphicsPolygonItem( pts );
  poly->setBrush( QBrush( Qt::black ) );
  poly->setPos( x, y );
  poly->setZValue( ZVALUE_LINE_PT );
  poly->show();
  // this -> addItem( poly );
  // line_points.push_back( std::pair< QGraphicsItem*, ThLine2D*>(poly,cur_line) );
  return poly;
}

void
PlotPainter::rotatePenBrush( QPen & pen, double x1, double y1, double x2, double y2 )
{
  QBrush brush = pen.brush();
  QMatrix m;
  double angle = atan2( y2-y1, x2-x1 ) - M_PI/2 ;
  m.translate( x1-10*cos(angle),  y1-10*sin(angle));
  m.rotate( angle * 180 / M_PI );
  brush.setMatrix( m );
  pen.setBrush( brush );
}

QGraphicsLineItem *
PlotPainter::createLineSegment( double x1, double y1, double x2, double y2, const QPen & pen )
{
  DBG_CHECK("createLineSegment %.2f %.2f - %.2f %.2f\n", x1, y1, x2, y2 );
  QLineF linef( x1, y1, x2, y2 );
  QGraphicsLineItem * line = new QGraphicsLineItem( linef );
  line->setPen( pen );
  line->setZValue( ZVALUE_LINE );
  line->show();
  // this -> addItem( line );
  // line_segments.push_back( std::pair< QGraphicsLineItem*, ThLine2D *>( line, cur_line ) );
  return line;
}

void
PlotPainter::createAreaSegmentTo( double x, double y, 
                                  ThLinePoint * prev, Therion::AreaType type,
                                  QGraphicsItem ** ppt, QGraphicsLineItem ** psgm )
{
  DBG_CHECK("createAreaSegmentTo() %.2f %.2f \n", x, y );
  // end-point
  if ( ppt ) {
    QGraphicsItem * pt = NULL;
    if ( type == Therion::THA_WATER ) {
      pt = createRoundAreaPoint( x, y, Qt::blue );
    } else if ( type == Therion::THA_SNOW || type == Therion::THA_ICE ) {
      pt = createRoundAreaPoint( x, y, Qt::gray );
    } else if ( type == Therion::THA_USER ) {
      pt = createRoundAreaPoint( x, y, Qt::green );
    } else {
      pt = createRoundAreaPoint( x, y, Qt::black );
    }
    *ppt = pt;
  }

  // area-segment
  if ( psgm ) {
    QGraphicsLineItem * sgm = NULL;
    if ( prev ) {
      QColor color;
      QPen pen;
      QBrush brush;
      icon->AreaStyle( type, color, pen, brush );
      // if ( type == Therion::THA_ICE ) {
      //   rotatePenBrush( pen, prev->x, prev->y, x, y );
      // }
      sgm = createAreaSegment( prev->x, prev->y, x, y, pen );
    }
    *psgm = sgm;
  }
}

QGraphicsLineItem *
PlotPainter::createAreaSegment( double x1, double y1, double x2, double y2, const QPen & pen )
{
  DBG_CHECK("createAreaSegment %.2f %.2f - %.2f %.2f\n", x1, y1, x2, y2 );
  QLineF linef( x1, y1, x2, y2 );
  QGraphicsLineItem * line = new QGraphicsLineItem( linef );
  line->setPen( pen );
  line->setZValue( ZVALUE_LINE );
  line->show();
  return line;
}

QGraphicsEllipseItem *
PlotPainter::createRoundAreaPoint( double x, double y, const QColor & color )
{
  DBG_CHECK("createRoundAreaPoint %.2f %.2f\n", x, y );
  QGraphicsEllipseItem * ell = new QGraphicsEllipseItem( x-2, y-2, 4, 4 );
  ell->setBrush( QBrush( color ) );
  ell->setZValue( ZVALUE_LINE_PT );
  ell->show();
  return ell;
}

QGraphicsItem *
PlotPainter::createPoint( ThPoint2D * pt, double s )
{
  double x = pt->x * s;
  double y = pt->y * s;
  Therion::PointType type = pt->type();
  int orient = pt->orientation() * ORIENTATION_UNITS;

#ifdef HAS_THP_ICONS
  int thp_off = -12;
  QGraphicsPixmapItem * item =
    new QGraphicsPixmapItem( icon->ThpPixmap( type ) );
  item -> setOffset( thp_off, thp_off );
#else
  // const QPolygon & poly = icon->ThpSymbol1( type );
  // QGraphicsPolygonItem * item = new QGraphicsPolygonItem( poly );
  const QPainterPath & path = icon->ThpSymbol2( type );
  QGraphicsPathItem * item = new QGraphicsPathItem( path );
  item->setBrush( icon->ThpBrush( type ) );
  item->setPen( icon->ThpPen( type ) );
#endif

  switch ( type ) {
    case Therion::THP_AIR_DRAUGHT:
    case Therion::THP_ENTRANCE:
    case Therion::THP_WATER_FLOW:
      item -> rotate( orient );
      break;
    default:
      break;
  }
  pt->item = item;
  item -> setPos( x, y );
  item -> setZValue( ZVALUE_POINT );
  item -> show();
  // this -> addItem( item );
  // g_points.push_back( std::pair< QGraphicsItem*, ThPoint2D*>( item, pt ) );
  return item;
}

QGraphicsItem *
PlotPainter::createLabelPoint( ThPoint2D * pt, double s, const char * text, int orient )
{
  double x = pt->x * s;
  double y = pt->y * s;
  QString t("ABC");
  // if ( option ) {
  //   const char * ch1 = strchr( option, '"' );
  //   const char * ch2 = strrchr( option, '"' );
  //   if ( ch1 && ch2 && ch1 != ch2 ) {
  //     t = ch1+1;
  //     t.truncate( ch2 - ch1 - 1 );
  //   }
  // }
  if ( text && strlen(text) > 0 ) t = text;
  // QGraphicsSimpleTextItem * txt = this ->addSimpleText( t );
  QGraphicsSimpleTextItem * txt = new QGraphicsSimpleTextItem( t );
  txt->rotate( orient * ORIENTATION_UNITS );
  txt->setPos( x, y );
  txt->setBrush( Qt::black );
  txt->show();
  pt->item = txt;
  return txt;
}
