/** @file PlotPainter.cpp
 *
 * @author marco corvi
 * @date feb 2011
 *
 * @brief plot therion area
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#include <QColor>
#include <QPen>
#include <QGraphicsItem>
#include <QGraphicsPolygonItem>
#include <QGraphicsEllipseItem>
#include <QGraphicsLineItem>
#include <QGraphicsPathItem>

#include "ArgCheck.h"
#include "PlotPainter.h"
#include "PlotZValue.h"
#include "TherionLine.h"


void
PlotPainter::rotatePenBrush( QPen & pen, double x1, double y1, double x2, double y2 )
{
  QBrush brush = pen.brush();
  QMatrix m;
  double angle = atan2( y2-y1, x2-x1 ) - M_PI/2 ;
  m.translate( x1-10*cos(angle),  y1-10*sin(angle));
  m.rotate( angle * 180 / M_PI );
  brush.setMatrix( m );
  pen.setBrush( brush );
}

PlotLineItem *
PlotPainter::makeLine( ThLinePoint * from, ThLinePoint * to, const QPen & pen )
{
  double x1, y1, x2, y2;
  double x1c, y1c, x2c, y2c;
  ThLine2D::computeCP( from, to, x1, y1, x1c, y1c, x2, y2, x2c, y2c );
  return new PlotLineItem( x1, y1, x1c, y1c, x2c, y2c, x2, y2, pen );
}


// ------------------------------------------------------------------------------
// LINE

QGraphicsItem * 
PlotPainter::createLinePoint( ThLinePoint * line_pt,  ThLinePoint * prev, Therion::LineType type )
{
  double x = line_pt -> x;
  double y = line_pt -> y;
  QColor color = icon->LinePointColor( type );
  QGraphicsItem * pt = NULL;
  if ( type == Therion::THL_ARROW ) {
    if ( prev == NULL ) {
      pt = createRoundLinePoint( x, y, color );
    } else {
      int k = 0;
      double dx = x - prev->x;
      double dy = y - prev->y;
      if ( fabs(dx) > 2*fabs(dy) ) {
        k = (dx > 0)? 2 : 6;
      } else if ( fabs(dy) > 2*fabs(dx) ) { 
        k = (dy > 0)? 4 : 0;
      } else if ( dx > 0 ) {
        k = (dy > 0)? 3 : 1;
      } else {
        k = (dy > 0)? 5 : 7;
      }
      pt = createPolyPoint( x, y, icon->ArrowEnd(k) );
    }
  } else {
    pt = createRoundLinePoint( x, y, color );
  }
  return pt;
}

PlotLineItem *
PlotPainter::createLineSegment( ThLinePoint * from, ThLinePoint * to, Therion::LineType type )
{
  PlotLineItem * sgm = NULL;
  if ( from != NULL ) {
    QColor color;
    QPen pen;
    icon->LineStyle( type, pen, color );
    if ( type == Therion::THL_PIT || type == Therion::THL_CHIMNEY ) {
      rotatePenBrush( pen, from->x, from->y, to->x, to->y );
    }
    sgm = makeLine( from, to, pen );
  }
  return sgm;
}

void
PlotPainter::createLineSegmentTo( ThLinePoint * prev,
                                  ThLinePoint * cur,
                                  Therion::LineType type,
                                  QGraphicsItem ** ppt, PlotLineItem ** psgm )
{
  if ( ppt ) { // create end-point
    *ppt = createLinePoint( cur, prev, type );
  }
  if ( psgm ) { // create line-segment
    * psgm = createLineSegment( prev, cur, type );
  }
}

// used for LINE points
QGraphicsEllipseItem *
PlotPainter::createRoundLinePoint( double x, double y, const QColor & color )
{
  DBG_CHECK("createRoundLinePoint %.2f %.2f\n", x, y );
  QGraphicsEllipseItem * ell = new QGraphicsEllipseItem( x-2, y-2, 4, 4 );
  ell->setPen(  color );
  ell->setBrush( QBrush( color ) );
  ell->setZValue( ZVALUE_LINE_PT );
  ell->show();
  // this -> addItem( ell );
  // line_points.push_back( std::pair< QGraphicsItem*, ThLine2D*>(ell,cur_line) );
  return ell;
}

// used for ARROW heads
QGraphicsItem *
PlotPainter::createPolyPoint( double x, double y, const QPolygon & pts )
{
  QGraphicsPolygonItem * poly = new QGraphicsPolygonItem( pts );
  poly->setBrush( QBrush( Qt::black ) );
  poly->setPos( x, y );
  poly->setZValue( ZVALUE_LINE_PT );
  poly->show();
  return poly;
}

//------------------------------------------------------------------------------
// POINT

QGraphicsItem *
PlotPainter::createPoint( ThPoint2D * pt, double s )
{
  double x = pt->x * s;
  double y = pt->y * s;
  Therion::PointType type = pt->type();
  int orient = pt->orientation() * ORIENTATION_UNITS;

#ifdef HAS_THP_ICONS
  int thp_off = -12;
  QGraphicsPixmapItem * item =
    new QGraphicsPixmapItem( icon->ThpPixmap( type ) );
  item -> setOffset( thp_off, thp_off );
#else
  // const QPolygon & poly = icon->ThpSymbol1( type );
  // QGraphicsPolygonItem * item = new QGraphicsPolygonItem( poly );
  const QPainterPath & path = icon->ThpSymbol2( type );
  QGraphicsPathItem * item = new QGraphicsPathItem( path );
  item->setBrush( icon->ThpBrush( type ) );
  item->setPen( icon->ThpPen( type ) );
#endif

  switch ( type ) {
    case Therion::THP_AIR_DRAUGHT:
    case Therion::THP_ENTRANCE:
    case Therion::THP_WATER_FLOW:
      item -> rotate( orient );
      break;
    default:
      break;
  }
  pt->item = item;
  item -> setPos( x, y );
  item -> setZValue( ZVALUE_POINT );
  item -> show();
  // this -> addItem( item );
  // g_points.push_back( std::pair< QGraphicsItem*, ThPoint2D*>( item, pt ) );
  return item;
}

QGraphicsItem *
PlotPainter::createLabelPoint( ThPoint2D * pt, double s, const char * text, int orient )
{
  double x = pt->x * s;
  double y = pt->y * s;
  QString t("ABC");
  // if ( option ) {
  //   const char * ch1 = strchr( option, '"' );
  //   const char * ch2 = strrchr( option, '"' );
  //   if ( ch1 && ch2 && ch1 != ch2 ) {
  //     t = ch1+1;
  //     t.truncate( ch2 - ch1 - 1 );
  //   }
  // }
  if ( text && strlen(text) > 0 ) t = text;
  // QGraphicsSimpleTextItem * txt = this ->addSimpleText( t );
  QGraphicsSimpleTextItem * txt = new QGraphicsSimpleTextItem( t );
  txt->rotate( orient * ORIENTATION_UNITS );
  txt->setPos( x, y );
  txt->setBrush( Qt::black );
  txt->show();
  pt->item = txt;
  return txt;
}

// ----------------------------------------------------------------------
// AREA

QGraphicsItem *
PlotPainter::createAreaPoint( ThLinePoint * at, Therion::AreaType type )
{
  if ( type == Therion::THA_WATER ) {
    return createRoundAreaPoint( at->x, at->y, Qt::blue );
  } else if ( type == Therion::THA_SNOW || type == Therion::THA_ICE ) {
    return createRoundAreaPoint( at->x, at->y, Qt::gray );
  } else if ( type == Therion::THA_USER ) {
    return createRoundAreaPoint( at->x, at->y, Qt::green );
  } else {
    return createRoundAreaPoint( at->x, at->y, Qt::black );
  }
  return NULL;
}

PlotLineItem *
PlotPainter::createAreaSegment( ThLinePoint * from, ThLinePoint * to,
                                Therion::AreaType type )
{
  if ( from ) {
    QColor color;
    QPen pen;
    QBrush brush;
    icon->AreaStyle( type, color, pen, brush );
    // if ( type == Therion::THA_ICE ) {
    //   rotatePenBrush( pen, prev->x, prev->y, x, y );
    // }
    return makeLine( from, to, pen ); 
  }
  return NULL;
}

void
PlotPainter::createAreaSegmentTo( ThLinePoint * from, ThLinePoint * to,
                                  Therion::AreaType type,
                                  QGraphicsItem ** ppt, PlotLineItem ** psgm )
{
  if ( ppt ) {
    *ppt = createAreaPoint( to, type );
  }
  if ( psgm ) {
    *psgm = createAreaSegment( from, to, type );
  }
}


QGraphicsEllipseItem *
PlotPainter::createRoundAreaPoint( double x, double y, const QColor & color )
{
  DBG_CHECK("createRoundAreaPoint %.2f %.2f\n", x, y );
  QGraphicsEllipseItem * ell = new QGraphicsEllipseItem( x-2, y-2, 4, 4 );
  ell->setBrush( QBrush( color ) );
  ell->setZValue( ZVALUE_LINE_PT );
  ell->show();
  return ell;
}
