/** @file PlotTh2Import.cpp
 *
 * @author marco corvi
 * @date march 2011
 * 
 * @brief Therion th2 import
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <vector>
#include <map>
#include <string>

#include "PlotTh2Import.h"
#include "PlotCanvas.h"
#include "PlotCanvasScene.h"

#include "ThPointType.h"
#include "ThLineType.h"
#include "ThAreaType.h"

enum Th2LineType {
  TH2_LINE_NONE,
  TH2_LINE_SCRAP,
  TH2_LINE_AREA,
  TH2_LINE_LINE,
  TH2_LINE_POINT,
  TH2_LINE_ENDSCRAP,
  TH2_LINE_ENDAREA,
  TH2_LINE_ENDLINE
};

PlotTh2Import::PlotTh2Import( QTshotWidget * p )
  : parent( p ) 
  , pos( NULL )
  , fp( NULL )
{
  line =(char *)malloc( 1024 * sizeof(char) );
}

PlotTh2Import::~PlotTh2Import()
{
  if ( line ) free( line );
}

const char * 
PlotTh2Import::nextToken() 
{
  const char * ret = NULL;
restart:
  while ( pos && *pos && isspace(*pos) ) ++pos;
  if ( *pos ) {
    if ( *pos == '#' ) { // comment
      return NULL;
    }
    if ( pos[0] == '\\' && pos[1] == '\n' ) { // continuation
      memset( line, 0 , 1024 );
      n = 0;
      if ( getline( &line, &n, fp ) == -1 ) return NULL;
      pos = line;
      goto restart;
    }
    ret = pos;
    while ( *pos && ! isspace(*pos) ) ++pos;
    if ( *pos != 0 ) {
      *pos = 0;
      ++pos;
    }
  }
  return ret;
}

int
PlotTh2Import::numberOfTokens()
{
  int nr = 0;
  const char * ch = pos;
  while ( ch && *ch && *ch != '\n' ) {
    while ( *ch && isspace(*ch) ) ++ch;
    if ( *ch ) {
      ++ nr;
      while ( *ch && ! isspace(*ch) ) ++ch;
    }
  }
  return nr;
}
  

bool
PlotTh2Import::isAreaBorder( const char * prefix, size_t len )
{
  const char * ch = pos;
  while ( ch && *ch && isspace(*ch) ) ++ch;
  return strncmp( ch, "-id", 3 ) == 0 && strncmp( ch+4, prefix, len ) == 0;
}


PlotCanvas *
PlotTh2Import::loadTh2File( const char * filename ) 
{
  size_t len = strlen( filename );
  if ( len <= 4 ) {
    fprintf(stderr, "Can import only th2 files (len %d)\n", len);
    return NULL;
  }
  if ( strncmp( filename+len-4, ".th2", 4 ) != 0 ) {
    fprintf(stderr, "Can import only th2 files\n");
    return NULL;
  }
  double x, y;
  const char * key;
  const char * value;
  fp = fopen( filename, "r" );
  if ( fp == NULL ) return NULL;
  bool in_scrap = false;
  bool in_line  = false;
  bool in_area  = false;
  bool in_point = false;
  ThLinePoint * prev_line_pt = NULL;
  int xoff = 0;
  int yoff = 0;

  int mode = -1;
  PlotCanvas * canvas = NULL;
  PlotStatus * status = NULL;
  PlotCanvasScene * scene = NULL;
  TherionScrap * scrap = NULL;
  // TODO plot->setCanvas( ... )
  Therion::PointType pt_type = Therion::THP_PLACEMARK;
  Therion::LineType  ln_type = Therion::THL_PLACEMARK;
  Therion::AreaType  ar_type = Therion::THA_PLACEMARK;
  const char * pt_subtype;
  const char * ln_subtype;
  std::vector< std::pair< double, double > > area_line;
  std::map< std::string, std::string > line_token;
  std::map< std::string, std::string > area_token;
  std::string area_prefix;
  size_t area_prefix_len = 0;

  char plot_name[32];
  strcpy( plot_name, "plot" );
  {
    char * ch1 = strrchr( filename, '/' );
    if ( ch1 == NULL ) {
      len -= 4;
      if ( len >= 32 ) len = 32 - 1;
      memset( plot_name, 0, 32 );
      memcpy( plot_name, filename, len );
      // fprintf(stderr, "ch1 == NULL plot_name %s filename %s\n", plot_name, filename );
    } else if ( (int)len > (ch1-filename)+5 ) {
      len -= (ch1-filename)+5;
      if ( len >= 32 ) len = 32 - 1;
      memset( plot_name, 0, 32 );
      memcpy( plot_name, ch1+1, len );
      // fprintf(stderr, "ch1 != NULL plot_name %s filename %s\n", plot_name, filename );
    }
  }
  // fprintf(stderr, "loadTh2File ... filename %s \n", filename );
  memset( line, 0 , 1024 );
  n = 1024;
  while ( getline( &line, &n, fp ) != -1 ) {
    if ( strncmp( line, "#QTopo", 6 ) != 0 ) continue;
    const char * ch = strstr( line, "offset " );
    if ( ch ) {
      sscanf( ch+7, "%d %d", &xoff, &yoff );
      break;
    }
  }
  while ( getline( &line, &n, fp ) != -1 ) {
    Th2LineType type = TH2_LINE_NONE;
    ThPoint2D * pt = NULL;
    // ThPoint2D * ln = NULL;
    // ThPoint2D * ar = NULL;
    pos = line;
    const char * token = NULL;
    while ( (token = nextToken() ) != NULL ) {
      if ( strcmp( token, "encoding" ) == 0 ) break;
      switch ( type ) {
      case TH2_LINE_NONE:
        if ( ! in_scrap ) {
          if ( strcmp( token, "scrap" ) == 0 ) {
            const char * name = nextToken(); // scrap name
            const char * key  = nextToken(); key = key;
            const char * proj = nextToken();
            std::ostringstream oss;
            oss << name << "-area";
            area_prefix = oss.str();
            area_prefix_len = area_prefix.size();
            if ( canvas == NULL ) {
              if ( strncmp( proj, "plan", 4 ) == 0 ) {
                mode = MODE_PLAN;
              } else if ( strncmp( proj, "ext", 3 ) == 0 ) {
                mode = MODE_EXT;
              } 
              if ( mode >= 0 ) {
                // fprintf(stderr, "loadTh2File: new canvas: scrap id %s plot %s\n", name, plot_name );
                // FIXME here may be a problem because the scene needs x,y, w, h 
                canvas = parent->openPlot( mode, plot_name, name );
                status = canvas->getStatus();
                scene  = canvas->getScene();
                scrap = status->getScrap();
                // xoff = status->getOffsetX();
                // yoff = status->getOffsetY();
                // fprintf(stderr, "loadTh2File: status offset %d %d \n", xoff, yoff );
              }
            } else {
              if ( ( mode == MODE_PLAN && strncmp( proj, "plan", 4 ) == 0 ) 
                || ( mode == MODE_EXT && strncmp( proj, "ext",  3 ) == 0 ) ) {
                // fprintf(stderr, "loadTh2File: new scrap id %s\n", name );
                canvas->doNewScrap( name );
                scrap = status->getScrap();
              } else {
                fprintf(stderr, "loadTh2File: ERROR incompatible scrap projections\n");
              }
            }
            // if ( status ) {
            //   fprintf(stderr, "loadTh2File: new scrap id %s\n", name );
            //   canvas->doNewScrap( name );
            //   scrap = status->getScrap();
            // }
            type = TH2_LINE_SCRAP;
            in_scrap = true;
          } else {
            fprintf(stderr, "loadTh2File: TODO out-of-scrap token \"%s\"\n", token );
          }
        } else { // in_scrap
          if ( strcmp( token, "endscrap" ) == 0 ) {
            type = TH2_LINE_ENDSCRAP;
            in_scrap = false;
            scrap = NULL;
          } else if ( !in_point && strcmp( token, "point" ) == 0 ) {
            type = TH2_LINE_POINT;
            const char * tx = nextToken();
            const char * ty = nextToken();
            const char * tt = nextToken();
            char * ch = strchr( tt, ':' );
            if ( ch ) {
              *ch = 0;
              pt_type = Therion::getPointType( tt );
              pt_subtype = Therion::getPointSubtype( pt_type, ch+1 );
            } else {
              pt_type = Therion::getPointType( tt );
              pt_subtype = NULL;
            }
            if ( pt_type == Therion::THP_STATION ) {
              x = atof( tx ) / TH_FACTOR - xoff;
              y = - atof( ty ) / TH_FACTOR -yoff;
            } else {
              x = atof( tx ) / TH_FACTOR * 4.0;
              y = - atof( ty ) / TH_FACTOR * 4.0;
            }
            // fprintf(stdout, "loadTh2File: createPoint %s %.2f %.2f\n", tt, x, y );
            pt = scene->insertPoint( x, y, pt_type );
            if ( pt_subtype )
              pt->setSubtype( pt_subtype );
            in_point = true;
          } else if ( !in_line && strcmp( token, "line" ) == 0 ) {
            const char * name = nextToken(); // line type 
            type = TH2_LINE_LINE;
            in_line = true;
            line_token.clear();
            if ( isAreaBorder( area_prefix.c_str(), area_prefix_len ) ) {
              const char * key = nextToken(); key = key;
              const char * id  = nextToken(); id = id;
              in_area = true;
            } else {
              char * ch = strchr( name, ':' );
              if ( ch ) {
                *ch = 0;
                ln_type = Therion::getLineType( name );
                ln_subtype = Therion::getLineSubtype( ln_type, ch+1 );
              } else {
                ln_type = Therion::getLineType( name );
                ln_subtype = NULL;
              }
            }
          } else if ( !in_area && strcmp( token, "area" ) == 0 ) {
            const char * name = nextToken(); // area type 
            ar_type = Therion::getAreaType( name );
            area_token.clear();
            // FIXME now must set the proper area type
            // FIXME ignore area lines
            for ( std::vector< std::pair< double, double > >::iterator 
                  it = area_line.begin(); it != area_line.end(); ++it ) {
              double x = it->first;
              double y = it->second;
              scene->insertAreaPoint( x, y, ar_type );
            }
            scene->endArea();
            type = TH2_LINE_AREA;
            in_area = true;
          } else if ( in_line && strcmp( token, "endline" ) == 0 ) {
            if ( in_area ) {
              in_area = false;
            } else {
              // fprintf(stderr, "TODO flush line_token\n");
              ThLine2D * line = scene->curLine();
              for ( std::map< std::string, std::string >::iterator it = line_token.begin();
                    it != line_token.end();
                    ++it ) {
                const std::string & key = it->first;
                const std::string & value = it->second;
                fprintf(stderr, "line token: %s %s \n", key.c_str(), value.c_str() );
                if ( key == "-subtype" ) {
                  line->setSubtype( value.c_str() );
                } else if ( key == "-close" ) {
                  if ( value == "on" ) {
                    size_t size = line->pointSize();
                    ThLinePoint * pt = line->operator[](0);
                    QGraphicsLineItem * sgm = NULL;
                    scene->createLineSegmentTo( pt, ln_type, size, NULL, &sgm );
                    line->closeLine( sgm );
                    // line->setClosed( true );
                  }
                } else if ( key == "-reverse" ) {
                  line->setReversed( (value == "on") );
                } else if ( key == "-clip" ) {
                  line->setClipped( (value == "on") );
                } else if ( key == "-visibility" ) {
                  line->setVisible(  (value == "on") );
                } else {
                  switch ( ln_type ) {
                    case Therion::THL_ARROW:
                    case Therion::THL_SLOPE:
                    case Therion::THL_CONTOUR:
                      if ( key == Therion::LineSpecialOptionName[ ln_type ] ) {
                        const char * val = Therion::getLineSpecialOption( ln_type, value.c_str() );
                        if ( val ) line->setSpecialString( val );
                      } else {
                        fprintf(stderr, "line[1] unhandled token: %s %s \n", key.c_str(), value.c_str() );
                      }
                      break;
                    case Therion::THL_PIT:
                    case Therion::THL_WALL:
                      if ( key == Therion::LineSpecialOptionName[ ln_type ] ) {
                        line->setSpecialValue( atoi(value.c_str()) );
                      } else {
                        fprintf(stderr, "line[2] unhandled token: %s %s \n", key.c_str(), value.c_str() );
                      }
                      break;
                    default:
                      fprintf(stderr, "line[3] unhandled token: %s %s \n", key.c_str(), value.c_str() );
                  }
                }
              }
              line_token.clear();
              scene->endLine();
            }
            type = TH2_LINE_ENDLINE;
            in_line = false;
            prev_line_pt = NULL;
          } else if ( in_area && strcmp( token, "endarea" ) == 0 ) {
            scene->endLine();
            for ( std::map< std::string, std::string >::iterator it = area_token.begin();
                  it != area_token.end();
                  ++it ) {
              fprintf(stderr, "TODO[4] area token: %s %s \n", it->first.c_str(), it->second.c_str() );
            }
            area_token.clear();
            type = TH2_LINE_ENDAREA;
            in_area = false;
          } else { // line points / area line-ids
            if ( in_line ) {
              if ( numberOfTokens() >= 5 ) {
                double x2 = atof( token ) / TH_FACTOR * 4.0;
                token = nextToken();
                double y2 = - atof( token ) / TH_FACTOR * 4.0;
                token = nextToken();
                double x1 = atof( token ) / TH_FACTOR * 4.0;
                token = nextToken();
                double y1 = - atof( token ) / TH_FACTOR * 4.0;
                token = nextToken();
                x = atof( token ) / TH_FACTOR * 4.0;
                token = nextToken();
                y = - atof( token ) / TH_FACTOR * 4.0;
                if ( in_area ) {
                  // FIXME handle control points for the area border
                  area_line.push_back( std::pair<double,double>(x,y) );
                } else {
                  if ( prev_line_pt ) {
                    scene->setControlPoint2(x2, y2, prev_line_pt );
                  }
                  prev_line_pt = scene->insertLinePoint( x, y, ln_type );
                  scene->setControlPoint1(x1, y1, prev_line_pt );
                }
              } else {
                x = atof( token ) / TH_FACTOR * 4.0;
                token = nextToken();
                y = - atof( token ) / TH_FACTOR * 4.0;
                if ( in_area ) {
                  area_line.push_back( std::pair<double,double>(x,y) );
                } else {
                  scene->insertLinePoint( x, y, ln_type );
                }
              }
            } else if ( in_area ) {
              // line id: ignore
            } else {
              fprintf(stderr, "loadTh2File: TODO token %s\n", token );
            }
          }
        }
        break;
      case TH2_LINE_SCRAP:
        key = token;
        value = nextToken();
        if ( strcmp(key, "-scale") == 0 ) {
          while ( nextToken() ) ;
        } else {
          fprintf(stdout, "loadTh2File: TODO scrap token: %s = %s\n", key, value );
        }
        break;
      case TH2_LINE_AREA:
        key = token;
        value = nextToken();
        // fprintf(stdout, "loadTh2File: area token: %s = %s\n", key, value );
        area_token[ key ] = value;
        break;
      case TH2_LINE_LINE:
        key = token;
        value = nextToken();
        // fprintf(stdout, "loadTh2File: line token: %s = %s\n", key, value );
        line_token[ key ] = value;
        break;
      case TH2_LINE_POINT:
        key   = token;
        value = nextToken();
        // fprintf(stdout, "loadTh2File: point token: %s = %s\n", key, value );
        if ( pt ) {
          if ( strcmp( key, "-name" ) == 0 ) {
            if ( pt_type == Therion::THP_STATION ) {
              pt->setText( value ); 
            } else {
              fprintf(stdout, "loadTh2File: TODO point station token: %s = %s\n", key, value );
            }
          } else if  ( strcmp( key, "-text" ) == 0 ) {
            pt->setText( value );
          } else if ( strncmp( key, "-orient", 7) == 0 ) {
            pt->setOrientation( atoi(value) );
          } else if ( strcmp( key, "-align" ) == 0 ) {
            const char * val = Therion::getPointAlign( value );
            pt->setAlign( val );
          } else if ( strcmp( key, "-scale" ) == 0 ) {
            const char * val = Therion::getPointScale( value );
            pt->setScale( val );
          } else {
            pt->appendOption( key, value );
          }
          // FIXME station comments
        }
        break;
      case TH2_LINE_ENDSCRAP:
      case TH2_LINE_ENDAREA:
      case TH2_LINE_ENDLINE:
        fprintf(stderr, "loadTh2File: TODO endxxx token \"%s\"\n", token );
        break;
      }
    }
    memset( line, 0 , 1024 );
    n = 1024;
    in_point = false;
  }
  return canvas;
}

#ifdef TEST

int main( int argc, char ** argv )
{
  if ( argc < 2 ) {
    fprintf(stderr, "Usage: %s th2_file\n", argv[0] );
    return 0;
  }
 
  PlotTh2Import th2;
  th2.loadTh2File( argv[1] );
  return 0;
}

#endif
