/** @file TherionLine.cpp
 *
 * @author marco corvi
 * @date aug. 2009
 *
 * @brief plot therion line
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#include "TherionLine.h"
#include "PlotLineItem.h"

void 
ThLinePoint::shift( double x0, double y0 )
{
  x  += x0;
  y  += y0;
  x1 += x0;
  y1 += y0;
  x2 += x0;
  y2 += y0;
  item->moveBy( x0*4, y0*4 );
}

void 
ThLinePoint::setControl1( double x10, double y10 )
{
  if ( fabs(x10-x) > 2.0 || fabs(y10-y) > 2.0 ) {
    x1 = x10;
    y1 = y10;
    has1 = true;
  }
}

void 
ThLinePoint::setControl2( double x20, double y20 )
{
  if ( fabs(x20-x) > 2.0 || fabs(y20-y) > 2.0 ) {
    x2 = x20;
    y2 = y20;
    has2 = true;
  }
}

void 
ThLinePoint::setControls( double x20, double y20 )
{
  if ( fabs(x20-x) > 2.0 || fabs(y20-y) > 2.0 ) {
    x1 = 2*x - x20;
    y1 = 2*y - y20;
    x2 = x20;
    y2 = y20;
    has1 = has2 = true;
    // fprintf(stderr, "setControls LP %.2f %.2f CP %.2f %.2f %.2f %.2f\n",
    //   x, y, x1, y1, x2, y2 );
  }
}


// -----------------------------------------------------------------------
// ThLineSegment

void
ThLineSegment::shift( double x0, double y0 )
{
  item->moveBy( x0*4, y0*4 );
}

// -----------------------------------------------------------------------
// ThLine2D

void 
ThLine2D::shift( double dx, double dy )
{
  for ( point_iterator it = pointBegin(), end = pointEnd(); it != end; ++it ) {
    (*it)->shift( dx, dy );
  }
  for ( segment_iterator it = segmentBegin(), end = segmentEnd(); it != end; ++it ) {
    (*it)->shift( dx, dy );
  }
}
  
bool 
ThLine2D::unclose()
{
  if ( closed ) { // drop last segment
    dropLastSegment(); // FIX-2
    closed = false;
  }
  return closed;
}
  
bool 
ThLine2D::closeLine( PlotLineItem * sgm_item )
{
  if ( ! closed ) {
    size_t s = pts.size();
    if ( s > 0 ) {
      insertSegment(s-1, 0, sgm_item);
      closed = true;
    }
  }
  return closed;
}
  
void 
ThLine2D::addLinePoint( ThLinePoint * p2, QGraphicsItem * pt_item, PlotLineItem * sgm_item )
{ 
  pts.push_back( p2 );
  p2->item = pt_item;
  if ( sgm_item ) {
    size_t s = pts.size();
    insertSegment(s-2, s-1, sgm_item);
  }
}


void 
ThLine2D::dropLastPoint()
{
  if ( pts.size() == 0 ) return;
  ThLinePoint * pt = pts.back();
  delete pt;
  pts.pop_back();
}

void 
ThLine2D::dropLastSegment()
{
  if ( sgms.size() == 0 ) return;
  ThLineSegment * sgm = sgms.back();
  delete sgm;
  sgms.pop_back(); // erase  from vector
}

void 
ThLine2D::insertSegment( size_t s1, size_t s2, PlotLineItem * sgm_item )
{
  ThLinePoint * p2 = pts[s1];
  ThLinePoint * p1 = pts[s2];
  ThLineSegment * sgm = new ThLineSegment( p1, p2, sgm_item );
  sgms.push_back( sgm );
}

  
void
ThLine2D::computeCP( ThLinePoint * from, ThLinePoint * to,
                        double & x1, double & y1, double & x1c, double & y1c,
                        double & x2, double & y2, double & x2c, double & y2c )
{
  x1 = from -> x;
  y1 = from -> y;
  x2 = to -> x;
  y2 = to -> y;
  if ( from->has2 ) {
    x1c = from -> x2;
    y1c = from -> y2;
    if ( to->has1 ) {
      x2c = to -> x1;
      y2c = to -> y1;
      // printf("MakeLine [FT] From %.2f %.2f %.2f %.2f To %.2f %.2f %.2f %.2f\n",
      //        x1, y1, x1c, y1c, x2, y2, x2c, y2c ); 
    } else {
      double dx   = x1 - x2;
      double dy   = y1 - y2;
      double dx1c = x1c - x1;
      double dy1c = y1c - y1;
      double f = 1.0 + (dx1c*dx + dy1c*dy)/(dx*dx + dy*dy);
      x2c = x1c - dx*f;
      y2c = y1c - dy*f;
      // printf("MakeLine [F] From %.2f %.2f %.2f %.2f To %.2f %.2f %.2f %.2f\n",
      //        x1, y1, x1c, y1c, x2, y2, x2c, y2c ); 
    }
  } else {
    if ( to->has1 ) {
      x2c = to -> x1;
      y2c = to -> y1; 
      double dx   = x2 - x1;
      double dy   = y2 - y1;
      double dx2c = x2c - x2;
      double dy2c = y2c - y2;
      double f = 1.0 + (dx2c*dx + dy2c*dy)/(dx*dx + dy*dy);
      x1c = x2c - dx*f;
      y1c = y2c - dy*f;
      // printf("MakeLine [T] From %.2f %.2f %.2f %.2f To %.2f %.2f %.2f %.2f\n",
      //        x1, y1, x1c, y1c, x2, y2, x2c, y2c ); 
    } else {
      double dx = (x2 - x1)/3;
      double dy = (y2 - y1)/3;
      x1c = x1 + dx;
      y1c = y1 + dy;
      x2c = x2 - dx;
      y2c = y2 - dy;
      // printf("MakeLine [-] From %.2f %.2f To %.2f %.2f\n", x1, y1, x2, y2 ); 
    }
  }
}


