/** @file TherionArea.h
 *
 * @author marco corvi
 * @date feb 2011
 *
 * @brief plot therion area
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#ifndef THERION_AREA_H
#define THERION_AREA_H

#include <math.h> // fabs
#include <vector>

#include "ThAreaType.h"
#include "TherionPoint.h"
#include "TherionLine.h"

class PlotAreaItem; // forward

/** plot therion area (contour)
 */
struct ThArea2D
{
  public:
    typedef std::vector< ThLineSegment * >::iterator segment_iterator;

    typedef std::vector< ThLinePoint * >::iterator point_iterator;
    typedef std::vector< ThLinePoint * >::const_iterator const_point_iterator;
 
  private:
    Therion::AreaType type;              //!< area type
    bool finished;                       //!< whether the area is finished (closed)
    bool visible;                        //!< whether the are border is visible
    std::vector< ThLinePoint * > pts;    //!< contour line points
    std::vector< ThLineSegment * > sgms; //!< contour line segments
    PlotAreaItem * item;                 //!< area polygon item
  
  public:
    /** cstr
     * @param t  line type (default THL_USER)
     */
    ThArea2D( Therion::AreaType t = Therion::THA_WATER )
      : type( t )
      , finished( false )
      , visible( true )
      , item( NULL )
    { 
      if ( type == Therion::THA_USER ) {
        visible = false;
      }
    }
  
    ~ThArea2D()
    {
      size_t k0 = pts.size();
      for ( size_t k=0; k<k0; ++k) delete pts[k];
      pts.clear();
      k0 = sgms.size();
      for ( size_t k=0; k<k0; ++k) delete sgms[k];
      sgms.clear();
    }

    bool isBorderVisible() const { return visible; }
    void setBorderVisible( bool v ) { visible = v; }

    Therion::AreaType Type() const { return type; }

    void setItem( PlotAreaItem * it ) { item = it; }
    PlotAreaItem * getItem() const { return item; }

    void shift( double dx, double dy );
  
    /** accessor: get the number of control points
     * @return the size of the vector of points
     */
    size_t Size() const { return pts.size(); }
    size_t SegmentSize() const { return sgms.size(); }

    ThLinePoint * LastPoint() { return pts.back(); }
    ThLineSegment * LastSegment() { return sgms.back(); }

    /** unclose the line
     */
    bool Unfinish();
  
    bool Finish( PlotLineItem * sgm_item );

    bool IsFinished() const { return finished; }
  
    /** check if the line is closed
     * @return true if the lin eis closed
     */
    bool isFinished() const { return finished; }

    /** append a point
     * @param x   point X coord.
     * @param y   point Y coord.
     * @return pointer to the last inserted point
     */
    void addAreaPoint( ThLinePoint * pt,
                       QGraphicsItem * pt_item,
                       PlotLineItem * sgm_item );

    /** erase the last point and segment
     */
    void DropLast() 
    { 
      DropLastPoint();
      DropLastSegment();
    }

  private:
    void DropLastPoint();
    
    void DropLastSegment();

    void InsertSegment( size_t s1, size_t s2, PlotLineItem * sgm_item );
  
  public:
    segment_iterator SegmentBegin() { return sgms.begin(); }
    segment_iterator SegmentEnd() { return sgms.end(); }

    point_iterator Begin() { return pts.begin(); }
    const_point_iterator Begin() const { return pts.begin(); }

    point_iterator End() { return pts.end(); }
    const_point_iterator End() const { return pts.end(); }

    ThLinePoint * operator[]( int k ) { return pts[k]; }

};

#endif


