/** @file TherionArea.h
 *
 * @author marco corvi
 * @date feb 2011
 *
 * @brief plot therion area
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#ifndef THERION_AREA_H
#define THERION_AREA_H

#include <math.h> // fabs
#include <vector>

#include "ThAreaType.h"
#include "TherionPoint.h"
#include "TherionLine.h"


/** plot therion area (contour)
 */
struct ThArea2D
{
  public:
    typedef std::vector< ThLineSegment * >::iterator segment_iterator;

    typedef std::vector< ThLinePoint * >::iterator point_iterator;
    typedef std::vector< ThLinePoint * >::const_iterator const_point_iterator;
 
  private:
    Therion::AreaType type;              //!< area type
    bool finished;                       //!< whether the area is finished (closed)
    bool visible;                        //!< whether the are border is visible
    std::vector< ThLinePoint * > pts;    //!< contour line points
    std::vector< ThLineSegment * > sgms; //!< contour line segments
    QGraphicsPolygonItem * item;         //!< area polygon item
  
  public:
    /** cstr
     * @param t  line type (default THL_USER)
     */
    ThArea2D( Therion::AreaType t = Therion::THA_WATER )
      : type( t )
      , finished( false )
      , visible( true )
      , item( NULL )
    { 
      if ( type == Therion::THA_USER ) {
        visible = false;
      }
    }
  
    ~ThArea2D()
    {
      size_t k0 = pts.size();
      for ( size_t k=0; k<k0; ++k) delete pts[k];
      pts.clear();
      k0 = sgms.size();
      for ( size_t k=0; k<k0; ++k) delete sgms[k];
      sgms.clear();
    }

    bool isBorderVisible() const { return visible; }
    void setBorderVisible( bool v ) { visible = v; }

    Therion::AreaType Type() const { return type; }

    void setItem( QGraphicsPolygonItem * it ) { item = it; }
    QGraphicsPolygonItem * getItem() const { return item; }

    void shift( double dx, double dy )
    {
      for ( point_iterator it = Begin(), end = End(); it != end; ++it ) {
        (*it)->shift( dx, dy );
      }
      for ( segment_iterator it = SegmentBegin(), end = SegmentEnd(); it != end; ++it ) {
        (*it)->shift( dx, dy );
      }
      if ( item ) item->moveBy( dx*4, dy*4 );
    }
  
    /** accessor: get the number of control points
     * @return the size of the vector of points
     */
    size_t Size() const { return pts.size(); }
    size_t SegmentSize() const { return sgms.size(); }

    ThLinePoint * LastPoint() { return pts.back(); }
    ThLineSegment * LastSegment() { return sgms.back(); }

    /** unclose the line
     */
    bool Unfinish()
    {
      if ( finished ) { // drop last segment
        DropLastSegment(); // FIX-2
        finished = false;
      }
      return finished;
    }
  
    bool Finish( QGraphicsLineItem * sgm_item )
    {
      if ( ! finished ) {
        size_t s = pts.size();
        if ( s > 0 ) {
          InsertSegment(s-1, 0, sgm_item);
          finished = true;
        }
      }
      return finished;
    }

    bool IsFinished() const { return finished; }
  
    /** check if the line is closed
     * @return true if the lin eis closed
     */
    bool isFinished() const { return finished; }

    /** append a point
     * @param x   point X coord.
     * @param y   point Y coord.
     * @return pointer to the last inserted point
     */
    ThLinePoint * addAreaPoint( double x, double y,
                                QGraphicsItem * pt_item,
                                QGraphicsLineItem * sgm_item )
    { 
      ThLinePoint * p2 = new ThLinePoint( x, y, pt_item );
      pts.push_back( p2 );
      if ( sgm_item ) {
        size_t s = pts.size();
        InsertSegment(s-2, s-1, sgm_item);
      }
      return p2;
    }

    /** erase the last point and segment
     */
    void DropLast() 
    { 
      DropLastPoint();
      DropLastSegment();
    }

  private:
    void DropLastPoint()
    {
      if ( pts.size() == 0 ) return;
      ThLinePoint * pt = pts.back();
      delete pt;
      pts.pop_back();
    }
    
    void DropLastSegment()
    {
      if ( sgms.size() == 0 ) return;
      ThLineSegment * sgm = sgms.back();
      delete sgm;
      sgms.pop_back(); // erase  from vector
    }

    void InsertSegment( size_t s1, size_t s2, QGraphicsLineItem * sgm_item )
    {
      ThLinePoint * p2 = pts[s1];
      ThLinePoint * p1 = pts[s2];
      ThLineSegment * sgm = new ThLineSegment( p1, p2, sgm_item );
      sgms.push_back( sgm );
    }

  
  public:
    segment_iterator SegmentBegin() { return sgms.begin(); }
    segment_iterator SegmentEnd() { return sgms.end(); }

    point_iterator Begin() { return pts.begin(); }
    const_point_iterator Begin() const { return pts.begin(); }

    point_iterator End() { return pts.end(); }
    const_point_iterator End() const { return pts.end(); }

    ThLinePoint * operator[]( int k ) { return pts[k]; }

};

#endif


