/** @file TherionScrap.h
 *
 * @author marco corvi
 * @date dec 2009
 *
 * @brief plot status
 * --------------------------------------------------------
 *  Copyright This sowftare is distributed under GPL-3.0 or later
 *  See the file COPYING.
 */
#ifndef THERION_SCRAP_H
#define THERION_SCRAP_H

#include <stdlib.h>

#include <string>
#include <vector>

#include <QGraphicsItemGroup>

#include "TherionPoint.h"
#include "TherionLine.h"
#include "TherionArea.h"

#ifdef HAS_POCKETTOPO
  #include "PTfile.h"
#endif

// class PlotCanvasScene; // forward
struct PlotStatus; 

struct TherionScrap
{
  private:
    std::string scrap_name; //!< scrap name
    std::vector< ThPoint2D * > th_points;
    std::vector< ThLine2D * >  th_lines;
    std::vector< ThArea2D * >  th_areas;
    PlotStatus * status;                   //!< plot status
    QGraphicsItemGroup * grey;               //!< group of the graphics items
    // PlotCanvasScene * canvas;
    // int    orientation;    //!< orientation 0..7 [units of 45 degrees]

  public:
    TherionScrap( PlotStatus * st, const char * name )
      : status( st )
      , grey( NULL )
      // , orientation( 0 )
    { 
      setScrapName( name );
    }

    ~TherionScrap()
    {
      clear();
    }
  
    /** get the "grey" group
     * @return the grey item group
     */
    QGraphicsItemGroup * getGrey() const { return grey; }

    /** set the "grey" group
     * @param g  the grey item group
     */
    void setGrey( QGraphicsItemGroup * g ) { grey = g; }

    /** check if the scrap has a certain name
     * @param name   name to check
     * @return true if the scrap has the given name
     */
    bool hasName( const char * name ) const { return scrap_name == name; }

    // void setScene( PlotCanvasScene * scene ) { canvas = scene; }

    const char * getScrapName() const { return scrap_name.c_str(); }

    void setScrapName( const char * name ) 
    {
      scrap_name = name;
      // FIXME replace " " by "_"
      int pos = -1;
      while ( (pos = scrap_name.find(' ', pos+1) ) >= 0 ) {
        scrap_name.replace( pos, 1, 1, '_');
      }
    }

    /** check if there are drawing items
     * @return true if there are drawing items
     */
    bool hasItems()
    {
      return th_points.size() > 0 || th_lines.size() > 0 || th_areas.size() > 0 ;
    }

    void clear()
    {
      size_t k0 = th_points.size();
      for ( size_t k=0; k<k0; ++k) delete th_points[k];
      th_points.clear();
      k0 = th_lines.size();
      for ( size_t k=0; k<k0; ++k) delete th_lines[k];
      th_lines.clear();
      k0 = th_areas.size();
      for ( size_t k=0; k<k0; ++k) delete th_areas[k];
      th_areas.clear();
    }

    void shiftItems( int dx, int dy )
    {
      // if ( do_debug ) 
      // fprintf(stderr, "shiftItems: lines %d points %d DX %d DY %d\n",
      //   th_lines.size(), th_points.size(), dx, dy );
      std::vector< ThPoint2D * >::iterator pend = th_points.end();
      for ( std::vector< ThPoint2D * >::iterator it = th_points.begin(); it != pend; ++it ) {
        (*it)->shift( dx, dy );
      }
      for ( std::vector< ThLine2D * >::iterator lit = th_lines.begin(), lend = th_lines.end(); lit != lend; ++lit ) {
        (*lit)->shift( dx, dy );
      }
      for ( std::vector< ThArea2D * >::iterator ait = th_areas.begin(), aend = th_areas.end(); ait != aend; ++ait ) {
        (*ait)->shift( dx, dy );
      }
    }

  #ifdef HAS_POCKETTOPO
    /** export drawing to a pockettopo file
     * @param drawing  PocketTopo drawing in the pockettopo file
     */
    void exportPocketTopoFile( PTdrawing & drawing, int offx, int offy );
  #endif

    size_t pointsSize() const { return th_points.size(); }
    std::vector< ThPoint2D * >::iterator pointsBegin() { return th_points.begin(); }
    std::vector< ThPoint2D * >::iterator pointsEnd()   { return th_points.end(); }
    void pointsAdd( ThPoint2D * pt ) { th_points.push_back( pt ); }
    void pointsErase( std::vector< ThPoint2D * >::iterator it ) { th_points.erase( it ); }
    ThPoint2D * pointsEraseLast()
    {
      ThPoint2D * pt = ( th_points.back() );
      th_points.pop_back();
      return pt;
    }
    ThPoint2D * pointsBack() { return th_points.back(); }

    size_t linesSize() const { return th_lines.size(); }
    std::vector< ThLine2D * >::iterator linesBegin() { return th_lines.begin(); }
    std::vector< ThLine2D * >::iterator linesEnd()   { return th_lines.end(); }
    void linesAdd( ThLine2D * line ) { th_lines.push_back( line ); }
    void linesErase( std::vector< ThLine2D * >::iterator it ) { th_lines.erase( it ); }
    ThLine2D * linesEraseLast()
    {
      ThLine2D * line = ( th_lines.back() );
      th_lines.pop_back();
      return line;
    }
    ThLine2D * linesBack() { return th_lines.back(); }


    size_t areasSize() const { return th_areas.size(); }
    std::vector< ThArea2D * >::iterator areasBegin() { return th_areas.begin(); }
    std::vector< ThArea2D * >::iterator areasEnd()   { return th_areas.end(); }
    void areasAdd( ThArea2D * area ) { th_areas.push_back( area ); }
    void areasErase( std::vector< ThArea2D * >::iterator it ) { th_areas.erase( it ); }
    ThArea2D * areasEraseLast()
    {
      ThArea2D * area = ( th_areas.back() );
      th_areas.pop_back();
      return area;
    }
    ThArea2D * areasBack() { return th_areas.back(); }
};

#endif

